import re
import subprocess


def git_remote():
    git_repo = subprocess.run(
        ["git", "remote", "get-url", "origin"], capture_output=True, text=True
    ).stdout.strip()
    return extract_repository_name(git_repo)


def extract_repository_name(repository_url):
    if not repository_url:
        return

    return re.search(r"([^/:]*/[^/]*?)(?:\.git)?$", repository_url).group(1)



import json
from abc import ABC
from abc import abstractmethod
from typing import Callable
from typing import Optional

from dbt_platform_helper.providers.cloudformation import CloudFormation
from dbt_platform_helper.providers.copilot import _normalise_secret_name
from dbt_platform_helper.providers.copilot import connect_to_addon_client_task
from dbt_platform_helper.providers.copilot import create_addon_client_task
from dbt_platform_helper.providers.copilot import get_postgres_admin_connection_string
from dbt_platform_helper.providers.ecs import ECS
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.secrets import Secrets
from dbt_platform_helper.providers.vpc import VpcProvider
from dbt_platform_helper.utils.application import Application


class ConduitECSStrategy(ABC):
    @abstractmethod
    def get_data(self):
        pass

    @abstractmethod
    def start_task(self, data_context: dict):
        pass

    @abstractmethod
    def exec_task(self, data_context: dict):
        pass


class TerraformConduitStrategy(ConduitECSStrategy):
    def __init__(
        self,
        clients,
        ecs_provider: ECS,
        application: Application,
        addon_name: str,
        addon_type: str,
        access: str,
        env: str,
        io: ClickIOProvider,
        vpc_provider: Callable,
        get_postgres_admin_connection_string: Callable,
    ):
        self.clients = clients
        self.ecs_provider = ecs_provider
        self.io = io
        self.vpc_provider = vpc_provider
        self.access = access
        self.addon_name = addon_name
        self.addon_type = addon_type
        self.application = application
        self.env = env
        self.get_postgres_admin_connection_string = get_postgres_admin_connection_string

    def get_data(self):
        self.io.info("Starting conduit in Terraform mode.")
        try:
            cluster_arn = self.ecs_provider.get_cluster_arn_by_name(
                f"{self.application.name}-{self.env}-cluster"
            )
        except:
            cluster_arn = self.ecs_provider.get_cluster_arn_by_name(
                f"{self.application.name}-{self.env}"
            )
        return {
            "cluster_arn": cluster_arn,
            "task_def_family": self._generate_container_name(),
            "vpc_name": self._resolve_vpc_name(),
            "addon_type": self.addon_type,
            "access": self.access,
        }

    def start_task(self, data_context: dict):

        environments = self.application.environments
        environment = environments.get(self.env)
        env_session = environment.session

        vpc_provider = self.vpc_provider(env_session)
        vpc_config = vpc_provider.get_vpc(
            self.application.name,
            self.env,
            data_context["vpc_name"],
        )

        postgres_admin_env_vars = None
        if data_context["addon_type"] == "postgres" and data_context["access"] == "admin":
            postgres_admin_env_vars = [
                {
                    "name": "CONNECTION_SECRET",
                    "value": self.get_postgres_admin_connection_string(
                        self.clients.get("ssm"),
                        f"/copilot/{self.application.name}/{self.env}/secrets/{_normalise_secret_name(self.addon_name)}",
                        self.application,
                        self.env,
                        self.addon_name,
                    ),
                },
            ]

        cluster_name = data_context["cluster_arn"].split("/")[-1]

        self.ecs_provider.start_ecs_task(
            cluster_name,
            self._generate_container_name(),
            data_context["task_def_family"],
            vpc_config,
            postgres_admin_env_vars,
        )

    def exec_task(self, data_context: dict):
        self.ecs_provider.exec_task(data_context["cluster_arn"], data_context["task_arns"][0])

    def _generate_container_name(self):
        return f"conduit-{self.addon_type}-{self.access}-{self.application.name}-{self.env}-{self.addon_name}"

    def _resolve_vpc_name(self):
        ssm_client = self.clients["ssm"]
        parameter_key = f"/platform/applications/{self.application.name}/environments/{self.env}"

        try:
            response = ssm_client.get_parameter(Name=parameter_key)["Parameter"]["Value"]
            return json.loads(response)["vpc_name"]
        except ssm_client.exceptions.ParameterNotFound:
            self.io.abort_with_error(
                f"Could not find AWS SSM parameter {parameter_key}. Please ensure your environment Terraform is up to date."
            )
        except KeyError:
            self.io.abort_with_error(
                f"The parameter {parameter_key} exists but does not contain the 'vpc_name' field. Please ensure your environment Terraform is up to date."
            )


class CopilotConduitStrategy(ConduitECSStrategy):
    def __init__(
        self,
        clients,
        ecs_provider: ECS,
        secrets_provider: Secrets,
        cloudformation_provider: CloudFormation,
        application: Application,
        addon_name: str,
        access: str,
        env: str,
        io: ClickIOProvider,
        connect_to_addon_client_task: Callable,
        create_addon_client_task: Callable,
    ):
        self.clients = clients
        self.cloudformation_provider = cloudformation_provider
        self.ecs_provider = ecs_provider
        self.secrets_provider = secrets_provider

        self.io = io
        self.access = access
        self.addon_name = addon_name
        self.application = application
        self.env = env
        self.connect_to_addon_client_task = connect_to_addon_client_task
        self.create_addon_client_task = create_addon_client_task

    def get_data(self):

        addon_type = self.secrets_provider.get_addon_type(self.addon_name)
        parameter_name = self.secrets_provider.get_parameter_name(
            addon_type, self.addon_name, self.access
        )
        task_name = self.ecs_provider.get_or_create_task_name(self.addon_name, parameter_name)

        return {
            "cluster_arn": self.ecs_provider.get_cluster_arn_by_copilot_tag(),
            "addon_type": addon_type,
            "task_def_family": f"copilot-{task_name}",
            "parameter_name": parameter_name,
            "task_name": task_name,
        }

    def start_task(self, data_context: dict):
        self.create_addon_client_task(
            self.clients["iam"],
            self.clients["ssm"],
            self.application,
            self.env,
            data_context["addon_type"],
            self.addon_name,
            data_context["task_name"],
            self.access,
        )

        self.io.info("Updating conduit task")
        self.cloudformation_provider.add_stack_delete_policy_to_task_role(data_context["task_name"])
        stack_name = self.cloudformation_provider.update_conduit_stack_resources(
            self.application.name,
            self.env,
            data_context["addon_type"],
            self.addon_name,
            data_context["task_name"],
            data_context["parameter_name"],
            self.access,
        )
        self.io.info("Waiting for conduit task update to complete...")
        self.cloudformation_provider.wait_for_cloudformation_to_reach_status(
            "stack_update_complete", stack_name
        )

    def exec_task(self, data_context: dict):
        self.connect_to_addon_client_task(
            self.clients["ecs"],
            self.application.name,
            self.env,
            data_context["cluster_arn"],
            data_context["task_name"],
        )


class ConduitStrategyFactory:

    @staticmethod
    def detect_mode(
        ecs_client,
        application,
        environment,
        addon_name: str,
        addon_type: str,
        access: str,
        io: ClickIOProvider,
    ) -> str:
        """Detect if Terraform-based conduit task definitions are present,
        otherwise default to Copilot mode."""
        paginator = ecs_client.get_paginator("list_task_definitions")
        prefix = f"conduit-{addon_type}-{access}-{application}-{environment}-{addon_name}"

        for page in paginator.paginate():
            for arn in page["taskDefinitionArns"]:
                if arn.split("/")[-1].startswith(prefix):
                    return "terraform"

        io.info("Defaulting to copilot mode.")
        return "copilot"

    @staticmethod
    def create_strategy(
        mode: str,
        clients,
        ecs_provider: ECS,
        secrets_provider: Secrets,
        cloudformation_provider: CloudFormation,
        application: Application,
        addon_name: str,
        addon_type: str,
        access: str,
        env: str,
        io: ClickIOProvider,
    ):

        if mode == "terraform":
            return TerraformConduitStrategy(
                clients,
                ecs_provider,
                application,
                addon_name,
                addon_type,
                access,
                env,
                io,
                vpc_provider=VpcProvider,
                get_postgres_admin_connection_string=get_postgres_admin_connection_string,
            )
        else:
            return CopilotConduitStrategy(
                clients,
                ecs_provider,
                secrets_provider,
                cloudformation_provider,
                application,
                addon_name,
                access,
                env,
                io,
                connect_to_addon_client_task=connect_to_addon_client_task,
                create_addon_client_task=create_addon_client_task,
            )


class Conduit:
    def __init__(
        self,
        application: Application,
        secrets_provider: Secrets,
        cloudformation_provider: CloudFormation,
        ecs_provider: ECS,
        io: ClickIOProvider = ClickIOProvider(),
        vpc_provider=VpcProvider,
        strategy_factory: Optional[ConduitStrategyFactory] = None,
    ):

        self.application = application
        self.secrets_provider = secrets_provider
        self.cloudformation_provider = cloudformation_provider
        self.ecs_provider = ecs_provider
        self.io = io
        self.vpc_provider = vpc_provider
        self.strategy_factory = strategy_factory or ConduitStrategyFactory()

    def start(self, env: str, addon_name: str, access: str = "read"):
        self.clients = self._initialise_clients(env)
        addon_type = self.secrets_provider.get_addon_type(addon_name)

        if (addon_type == "opensearch" or addon_type == "redis") and (access != "read"):
            access = "read"

        mode = self.strategy_factory.detect_mode(
            self.clients.get("ecs"),
            self.application.name,
            env,
            addon_name,
            addon_type,
            access,
            self.io,
        )

        strategy = self.strategy_factory.create_strategy(
            mode=mode,
            clients=self.clients,
            ecs_provider=self.ecs_provider,
            secrets_provider=self.secrets_provider,
            cloudformation_provider=self.cloudformation_provider,
            application=self.application,
            addon_name=addon_name,
            addon_type=addon_type,
            access=access,
            env=env,
            io=self.io,
        )

        data_context = strategy.get_data()

        data_context["task_arns"] = self.ecs_provider.get_ecs_task_arns(
            cluster=data_context["cluster_arn"], task_def_family=data_context["task_def_family"]
        )

        info_log = (
            f"Checking if a conduit ECS task is already running for:\n"
            f"  Addon Name : {addon_name}\n"
            f"  Addon Type : {addon_type}"
        )

        if addon_type == "postgres":
            info_log += f"\n  Access Level : {access}"

        self.io.info(info_log)

        if not data_context["task_arns"]:
            self.io.info("Creating conduit ECS task...")
            strategy.start_task(data_context)
            data_context["task_arns"] = self.ecs_provider.wait_for_task_to_register(
                data_context["cluster_arn"], data_context["task_def_family"]
            )
        else:
            self.io.info(f"Found a task already running: {data_context['task_arns'][0]}")

        self.io.info(f"Waiting for ECS Exec agent to become available on the conduit task...")
        self.ecs_provider.ecs_exec_is_available(
            data_context["cluster_arn"], data_context["task_arns"]
        )

        self.io.info("Connecting to conduit task...")
        strategy.exec_task(data_context)

    def _initialise_clients(self, env):
        return {
            "ecs": self.application.environments[env].session.client("ecs"),
            "iam": self.application.environments[env].session.client("iam"),
            "ssm": self.application.environments[env].session.client("ssm"),
        }



import click

from dbt_platform_helper.domain.codebase import Codebase
from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.parameter_store import ParameterStore
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.click import ClickDocOptGroup


@click.group(chain=True, cls=ClickDocOptGroup)
def codebase():
    """Codebase commands."""
    PlatformHelperVersioning().check_if_needs_update()


@codebase.command()
def prepare():
    """Sets up an application codebase for use within a DBT platform project."""
    try:
        Codebase(ParameterStore(get_aws_session_or_abort().client("ssm"))).prepare()
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@codebase.command()
@click.option("--app", help="AWS application name", required=True)
@click.option(
    "--with-images",
    help="List up to the last 10 images tagged for this codebase",
    default=False,
    is_flag=True,
)
def list(app, with_images):
    """List available codebases for the application."""
    try:
        Codebase(ParameterStore(get_aws_session_or_abort().client("ssm"))).list(app, with_images)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@codebase.command()
@click.option("--app", help="AWS application name", required=True)
@click.option(
    "--codebase",
    help="The codebase name as specified in the platform-config.yml file. This must be run from your codebase repository directory.",
    required=True,
)
@click.option("--commit", help="GitHub commit hash", required=True)
def build(app, codebase, commit):
    """Trigger a CodePipeline pipeline based build."""
    try:
        Codebase(ParameterStore(get_aws_session_or_abort().client("ssm"))).build(
            app, codebase, commit
        )
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@codebase.command()
@click.option("--app", help="AWS application name", required=True)
@click.option("--env", help="AWS Copilot environment", required=True)
@click.option(
    "--codebase",
    help="The codebase name as specified in the platform-config.yml file. This can be run from any directory.",
    required=True,
)
@click.option(
    "--tag",
    help="Git tag that has been built into an image. Typically a semantic version of the form 1.2.3 or v1.2.3.",
    required=False,
)
@click.option(
    "--branch",
    help="Git branch that has been built into an image.",
    required=False,
)
@click.option(
    "--commit",
    help="Git sha hash that has been built into an image.",
    required=False,
)
def deploy(
    app: str, env: str, codebase: str, commit: str = None, tag: str = None, branch: str = None
):

    try:
        Codebase(ParameterStore(get_aws_session_or_abort().client("ssm"))).deploy(
            app, env, codebase, commit, tag, branch
        )
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))



import ipaddress
import re
from typing import Callable

from schema import Optional
from schema import Or
from schema import Regex
from schema import Schema
from schema import SchemaError

from dbt_platform_helper.constants import PLATFORM_CONFIG_SCHEMA_VERSION
from dbt_platform_helper.domain.plans import PlanLoader

plan_manager = PlanLoader()
plan_manager.load()


class PlatformConfigSchema:
    @staticmethod
    def schema() -> Schema:
        return Schema(
            {
                "schema_version": PLATFORM_CONFIG_SCHEMA_VERSION,
                "application": str,
                Optional("deploy_repository"): str,
                "default_versions": PlatformConfigSchema.__default_versions_schema(),
                Optional("environments"): PlatformConfigSchema.__environments_schema(),
                Optional("codebase_pipelines"): PlatformConfigSchema.__codebase_pipelines_schema(),
                Optional(
                    "environment_pipelines"
                ): PlatformConfigSchema.__environment_pipelines_schema(),
                Optional("extensions"): {
                    str: Or(
                        PlatformConfigSchema.__alb_schema(),
                        PlatformConfigSchema.__monitoring_schema(),
                        PlatformConfigSchema.__opensearch_schema(),
                        PlatformConfigSchema.__postgres_schema(),
                        PlatformConfigSchema.__datadog_schema(),
                        PlatformConfigSchema.__prometheus_policy_schema(),
                        PlatformConfigSchema.__redis_schema(),
                        PlatformConfigSchema.__s3_bucket_schema(),
                        PlatformConfigSchema.__s3_bucket_policy_schema(),
                    )
                },
            }
        )

    @staticmethod
    def extension_schemas() -> dict:
        return {
            "alb": Schema(PlatformConfigSchema.__alb_schema()),
            "appconfig-ipfilter": PlatformConfigSchema.__no_configuration_required_schema(
                "appconfig-ipfilter"
            ),
            "opensearch": ConditionalOpensSearchSchema(PlatformConfigSchema.__opensearch_schema()),
            "postgres": Schema(PlatformConfigSchema.__postgres_schema()),
            "prometheus-policy": Schema(PlatformConfigSchema.__prometheus_policy_schema()),
            "redis": Schema(PlatformConfigSchema.__redis_schema()),
            "datadog": Schema(PlatformConfigSchema.__datadog_schema()),
            "s3": Schema(PlatformConfigSchema.__s3_bucket_schema()),
            "s3-policy": Schema(PlatformConfigSchema.__s3_bucket_policy_schema()),
            "subscription-filter": PlatformConfigSchema.__no_configuration_required_schema(
                "subscription-filter"
            ),
            # TODO: DBTP-1943: The next three are no longer relevant. Remove them.
            "monitoring": Schema(PlatformConfigSchema.__monitoring_schema()),
            "vpc": PlatformConfigSchema.__no_configuration_required_schema("vpc"),
            "xray": PlatformConfigSchema.__no_configuration_required_schema("xray"),
        }

    @staticmethod
    def __alb_schema() -> dict:
        _valid_alb_cache_policy = {
            "min_ttl": int,
            "max_ttl": int,
            "default_ttl": int,
            "cookies_config": Or("none", "whitelist", "allExcept", "all"),
            "header": Or("none", "whitelist"),
            "query_string_behavior": Or("none", "whitelist", "allExcept", "all"),
            Optional("cookie_list"): list,
            Optional("headers_list"): list,
            Optional("cache_policy_query_strings"): list,
        }

        _valid_alb_paths_definition = {
            Optional("default"): {
                "cache": str,
                "request": str,
            },
            Optional("additional"): [
                {
                    "path": str,
                    "cache": str,
                    "request": str,
                }
            ],
        }

        return {
            "type": "alb",
            Optional("environments"): {
                PlatformConfigSchema.__valid_environment_name(): Or(
                    {
                        Optional("additional_address_list"): [str],
                        Optional("allowed_methods"): [str],
                        Optional("cached_methods"): [str],
                        Optional("cdn_compress"): bool,
                        Optional("cdn_domains_list"): dict,
                        Optional("cdn_geo_locations"): [str],
                        Optional("cdn_geo_restriction_type"): str,
                        Optional("cdn_logging_bucket"): str,
                        Optional("cdn_logging_bucket_prefix"): str,
                        Optional("cdn_timeout_seconds"): int,
                        Optional("default_waf"): str,
                        Optional("domain_prefix"): str,
                        Optional("enable_logging"): bool,
                        Optional("forwarded_values_forward"): str,
                        Optional("forwarded_values_headers"): [str],
                        Optional("forwarded_values_query_string"): bool,
                        Optional("origin_protocol_policy"): str,
                        Optional("origin_ssl_protocols"): [str],
                        Optional("slack_alert_channel_alb_secret_rotation"): str,
                        Optional("viewer_certificate_minimum_protocol_version"): str,
                        Optional("viewer_certificate_ssl_support_method"): str,
                        Optional("viewer_protocol_policy"): str,
                        Optional("cache_policy"): dict({str: _valid_alb_cache_policy}),
                        Optional("origin_request_policy"): dict({str: {}}),
                        Optional("paths"): dict({str: _valid_alb_paths_definition}),
                        Optional("managed_ingress"): bool,
                    },
                    None,
                )
            },
        }

    @staticmethod
    def __codebase_pipelines_schema() -> dict:
        return {
            str: {
                "repository": str,
                Optional("slack_channel"): str,
                Optional("requires_image_build"): bool,
                Optional("additional_ecr_repository"): str,
                Optional("deploy_repository_branch"): str,
                "services": [{str: [str]}],
                Optional("pipelines"): [
                    Or(
                        {
                            "name": str,
                            "branch": PlatformConfigSchema.__valid_branch_name(),
                            "environments": [
                                {
                                    "name": str,
                                    Optional("requires_approval"): bool,
                                },
                            ],
                        },
                        {
                            "name": str,
                            "tag": bool,
                            "environments": [
                                {
                                    "name": str,
                                    Optional("requires_approval"): bool,
                                }
                            ],
                        },
                    ),
                ],
                Optional("cache_invalidation"): {
                    "domains": PlatformConfigSchema.__cache_invalidation_domains_schema(),
                },
            },
        }

    @staticmethod
    def __cache_invalidation_domains_schema() -> dict:
        return {str: {"paths": [str], "environment": str}}

    @staticmethod
    def __default_versions_schema() -> dict:
        return {
            "platform-helper": str,
        }

    @staticmethod
    def __environments_schema() -> dict:
        return {
            str: Or(
                None,
                {
                    Optional("accounts"): {
                        "deploy": {
                            "name": str,
                            "id": str,
                        },
                        "dns": {
                            "name": str,
                            "id": str,
                        },
                    },
                    # TODO: DBTP-1943: requires_approval is no longer relevant since we don't have AWS Copilot manage environment pipelines
                    Optional("requires_approval"): bool,
                    Optional("vpc"): str,
                    Optional("service-deployment-mode"): Or(
                        "copilot",
                        "dual-deploy-copilot-traffic",
                        "dual-deploy-platform-traffic",
                        "platform",
                    ),
                    Optional("network"): {
                        Optional("egress_rules"): {
                            str: {
                                "destination": Or(
                                    {
                                        "cidr_blocks": [str],
                                    },
                                    {
                                        "vpc_endpoints": bool,
                                    },
                                    {
                                        "aws_cidr_blocks": {
                                            "regions": [str],
                                            "services": [str],
                                        },
                                    },
                                ),
                                "protocol": Or("tcp", "udp", "-1"),  # "-1" means all protocols
                                "from_port": int,
                                "to_port": int,
                            },
                        },
                        Optional("vpc_endpoints"): {
                            str: {
                                "service_name": str,
                            },
                        },
                    },
                },
            )
        }

    @staticmethod
    def __environment_pipelines_schema() -> dict:
        _valid_environment_pipeline_specific_version_overrides = {
            Optional("platform-helper"): str,
        }

        return {
            str: {
                Optional("account"): str,
                Optional("branch", default="main"): PlatformConfigSchema.__valid_branch_name(),
                Optional("pipeline_to_trigger"): str,
                Optional("versions"): _valid_environment_pipeline_specific_version_overrides,
                "slack_channel": str,
                "trigger_on_push": bool,
                "environments": {
                    str: Or(
                        None,
                        {
                            Optional("accounts"): {
                                "deploy": {
                                    "name": str,
                                    "id": str,
                                },
                                "dns": {
                                    "name": str,
                                    "id": str,
                                },
                            },
                            Optional("requires_approval"): bool,
                            Optional(
                                "versions"
                            ): _valid_environment_pipeline_specific_version_overrides,
                            Optional("vpc"): str,
                        },
                    )
                },
            }
        }

    @staticmethod
    def __monitoring_schema() -> dict:
        return {
            "type": "monitoring",
            Optional("environments"): {
                PlatformConfigSchema.__valid_environment_name(): {
                    Optional("enable_ops_center"): bool,
                }
            },
        }

    @staticmethod
    def __opensearch_schema() -> dict:
        # TODO: DBTP-1943: Move to OpenSearch provider?
        _valid_opensearch_plans = Or(*plan_manager.get_plan_names("opensearch"))

        return {
            "type": "opensearch",
            Optional("environments"): {
                PlatformConfigSchema.__valid_environment_name(): {
                    Optional("engine"): str,
                    Optional("deletion_policy"): PlatformConfigSchema.__valid_deletion_policy(),
                    Optional("plan"): _valid_opensearch_plans,
                    Optional("volume_size"): int,
                    Optional("ebs_throughput"): int,
                    Optional("ebs_volume_type"): str,
                    Optional("instance"): str,
                    Optional("instances"): int,
                    Optional("master"): bool,
                    Optional("es_app_log_retention_in_days"): int,
                    Optional("index_slow_log_retention_in_days"): int,
                    Optional("audit_log_retention_in_days"): int,
                    Optional("search_slow_log_retention_in_days"): int,
                    Optional("password_special_characters"): str,
                    Optional("urlencode_password"): bool,
                }
            },
        }

    @staticmethod
    def __postgres_schema() -> dict:
        _valid_postgres_plans = Or(*plan_manager.get_plan_names("postgres"))
        _valid_postgres_version = Or(int, float)

        # TODO: DBTP-1943: Move to Postgres provider?
        _valid_postgres_storage_types = Or("gp2", "gp3", "io1", "io2")

        _valid_postgres_database_copy = {
            "from": PlatformConfigSchema.__valid_environment_name(),
            "to": PlatformConfigSchema.__valid_environment_name(),
            Optional("pipeline"): {Optional("schedule"): str},
        }

        return {
            "type": "postgres",
            Optional("version"): _valid_postgres_version,
            Optional("deletion_policy"): PlatformConfigSchema.__valid_postgres_deletion_policy(),
            Optional("environments"): {
                PlatformConfigSchema.__valid_environment_name(): {
                    Optional("apply_immediately"): bool,
                    Optional("plan"): _valid_postgres_plans,
                    Optional("version"): (Or(int, float)),
                    Optional("volume_size"): PlatformConfigSchema.is_integer_between(20, 10000),
                    Optional("iops"): PlatformConfigSchema.is_integer_between(1000, 9950),
                    Optional("snapshot_id"): str,
                    Optional(
                        "deletion_policy"
                    ): PlatformConfigSchema.__valid_postgres_deletion_policy(),
                    Optional("deletion_protection"): bool,
                    Optional("multi_az"): bool,
                    Optional("storage_type"): _valid_postgres_storage_types,
                }
            },
            Optional("database_copy"): [_valid_postgres_database_copy],
            Optional("objects"): [
                {
                    "key": str,
                    Optional("body"): str,
                }
            ],
        }

    @staticmethod
    def __prometheus_policy_schema() -> dict:
        return {
            "type": "prometheus-policy",
            Optional("services"): Or("__all__", [str]),
            Optional("environments"): {
                PlatformConfigSchema.__valid_environment_name(): {
                    "role_arn": str,
                }
            },
        }

    @staticmethod
    def __redis_schema() -> dict:
        _valid_redis_plans = Or(*plan_manager.get_plan_names("redis"))

        return {
            "type": "redis",
            Optional("environments"): {
                PlatformConfigSchema.__valid_environment_name(): {
                    Optional("plan"): _valid_redis_plans,
                    Optional("engine"): str,
                    Optional("replicas"): PlatformConfigSchema.is_integer_between(0, 5),
                    Optional("deletion_policy"): PlatformConfigSchema.__valid_deletion_policy(),
                    Optional("apply_immediately"): bool,
                    Optional("automatic_failover_enabled"): bool,
                    Optional("instance"): str,
                    Optional("multi_az_enabled"): bool,
                }
            },
        }

    @staticmethod
    def valid_s3_bucket_name(name: str):
        # TODO: DBTP-1943: This is a public method becasue that's what the test expect. Perhaps it belongs in an S3 provider?
        errors = []
        if not (2 < len(name) < 64):
            errors.append("Length must be between 3 and 63 characters inclusive.")

        if not re.match(r"^[a-z0-9].*[a-z0-9]$", name):
            errors.append("Names must start and end with 0-9 or a-z.")

        if not re.match(r"^[a-z0-9.-]*$", name):
            errors.append("Names can only contain the characters 0-9, a-z, '.' and '-'.")

        if ".." in name:
            errors.append("Names cannot contain two adjacent periods.")

        try:
            ipaddress.ip_address(name)
            errors.append("Names cannot be IP addresses.")
        except ValueError:
            pass

        for prefix in ("xn--", "sthree-"):
            if name.startswith(prefix):
                errors.append(f"Names cannot be prefixed '{prefix}'.")

        for suffix in ("-s3alias", "--ol-s3"):
            if name.endswith(suffix):
                errors.append(f"Names cannot be suffixed '{suffix}'.")

        if errors:
            # TODO: DBTP-1943: Raise suitable PlatformException?
            raise SchemaError(
                f"Bucket name '{name}' is invalid:\n" + "\n".join(f"  {e}" for e in errors)
            )

        return True

    @staticmethod
    def __datadog_schema() -> dict:
        return {
            "type": "datadog",
            Optional("environments"): {
                Optional(PlatformConfigSchema.__valid_environment_name()): {
                    "team_name": str,
                    Optional("contact_name"): str,
                    Optional("contact_email"): str,
                    Optional("contacts"): [
                        {
                            "name": str,
                            "type": str,
                            "contact": str,
                        }
                    ],
                    Optional("documentation_url"): str,
                    "services_to_monitor": dict,
                    Optional("description"): str,
                },
            },
        }

    @staticmethod
    def __s3_bucket_schema() -> dict:
        def _valid_s3_bucket_arn(key):
            return Regex(
                r"^arn:aws:s3::.*",
                error=f"{key} must contain a valid ARN for an S3 bucket",
            )

        _single_import_config = {
            Optional("source_kms_key_arn"): PlatformConfigSchema.__valid_kms_key_arn(
                "source_kms_key_arn"
            ),
            "source_bucket_arn": _valid_s3_bucket_arn("source_bucket_arn"),
            "worker_role_arn": PlatformConfigSchema.__valid_iam_role_arn("worker_role_arn"),
        }

        _valid_s3_data_migration = {
            Optional("import"): _single_import_config,
            Optional("import_sources"): [_single_import_config],
        }

        _valid_s3_bucket_retention_policy = Or(
            None,
            {
                "mode": Or("GOVERNANCE", "COMPLIANCE"),
                Or("days", "years", only_one=True): int,
            },
        )

        _valid_s3_bucket_lifecycle_rule = {
            Optional("filter_prefix"): str,
            "expiration_days": int,
            "enabled": bool,
        }

        _valid_s3_bucket_external_role_access = {
            "role_arn": PlatformConfigSchema.__valid_iam_role_arn("role_arn"),
            "read": bool,
            "write": bool,
            "cyber_sign_off_by": PlatformConfigSchema.__valid_dbt_email_address(
                "cyber_sign_off_by"
            ),
        }

        _valid_s3_bucket_external_role_access_name = Regex(
            r"^([a-z][a-zA-Z0-9_-]*)$",
            error="External role access block name {} is invalid: names must only contain lowercase alphanumeric characters separated by hypen or underscore",
        )

        return dict(
            {
                "type": "s3",
                Optional("objects"): [
                    {"key": str, Optional("body"): str, Optional("content_type"): str}
                ],
                Optional("readonly"): bool,
                Optional("serve_static_content"): bool,
                Optional("serve_static_param_name"): str,
                Optional("services"): Or("__all__", [str]),
                Optional("environments"): {
                    PlatformConfigSchema.__valid_environment_name(): {
                        "bucket_name": PlatformConfigSchema.valid_s3_bucket_name,
                        Optional("deletion_policy"): PlatformConfigSchema.__valid_deletion_policy(),
                        Optional("retention_policy"): _valid_s3_bucket_retention_policy,
                        Optional("versioning"): bool,
                        Optional("lifecycle_rules"): [_valid_s3_bucket_lifecycle_rule],
                        Optional("data_migration"): _valid_s3_data_migration,
                        Optional("external_role_access"): {
                            PlatformConfigSchema.__valid_schema_key(): _valid_s3_bucket_external_role_access
                        },
                        Optional("cross_environment_service_access"): {
                            PlatformConfigSchema.__valid_schema_key(): {
                                # Deprecated: We didn't implement cross application access, no service teams are asking for it.
                                # application should be removed once we can confirm that no-one is using it.
                                Optional("application"): str,
                                "environment": PlatformConfigSchema.__valid_environment_name(),
                                "account": str,
                                "service": str,
                                "read": bool,
                                "write": bool,
                                "cyber_sign_off_by": PlatformConfigSchema.__valid_dbt_email_address(
                                    "cyber_sign_off_by"
                                ),
                            }
                        },
                        Optional("managed_ingress"): bool,
                    },
                },
            }
        )

    @staticmethod
    def __s3_bucket_policy_schema() -> dict:
        return dict(
            {
                "type": "s3-policy",
                Optional("services"): Or("__all__", [str]),
                Optional("environments"): {
                    PlatformConfigSchema.__valid_environment_name(): {
                        "bucket_name": PlatformConfigSchema.valid_s3_bucket_name,
                    },
                },
            }
        )

    @staticmethod
    def string_matching_regex(regex_pattern: str) -> Callable:
        # TODO: DBTP-1943: public for the unit tests, not sure about testing what could be a private method. Perhaps it's covered by other tests anyway?
        def validate(string):
            if not re.match(regex_pattern, string):
                # TODO: DBTP-1943: Raise suitable PlatformException?
                raise SchemaError(
                    f"String '{string}' does not match the required pattern '{regex_pattern}'."
                )
            return string

        return validate

    @staticmethod
    def is_integer_between(lower_limit, upper_limit) -> Callable:
        # TODO: DBTP-1943: public for the unit tests, not sure about testing what could be a private method. Perhaps it's covered by other tests anyway?
        def validate(value):
            if isinstance(value, int) and lower_limit <= value <= upper_limit:
                return True
            # TODO: DBTP-1943: Raise suitable PlatformException?
            raise SchemaError(f"should be an integer between {lower_limit} and {upper_limit}")

        return validate

    @staticmethod
    def __valid_schema_key() -> Regex:
        return Regex(
            r"^([a-z][a-zA-Z0-9_-]*|\*)$",
            error="{} is invalid: must only contain lowercase alphanumeric characters separated by hyphen or underscore",
        )

    @staticmethod
    def __valid_branch_name() -> Callable:
        # TODO: DBTP-1943: Make this actually validate a git branch name properly; https://git-scm.com/docs/git-check-ref-format
        return PlatformConfigSchema.string_matching_regex(r"^((?!\*).)*(\*)?$")

    @staticmethod
    def __valid_deletion_policy() -> Or:
        return Or("Delete", "Retain")

    @staticmethod
    def __valid_postgres_deletion_policy() -> Or:
        return Or("Delete", "Retain", "Snapshot")

    @staticmethod
    def __valid_environment_name() -> Regex:
        return Regex(
            r"^([a-z][a-zA-Z0-9]*|\*)$",
            error="Environment name {} is invalid: names must only contain lowercase alphanumeric characters, or be the '*' default environment",
            # For values the "error" parameter works and outputs the custom text. For keys the custom text doesn't get reported in the exception for some reason.
        )

    @staticmethod
    def __valid_kms_key_arn(key) -> Regex:
        return Regex(
            r"^arn:aws:kms:.*:\d{12}:(key|alias).*",
            error=f"{key} must contain a valid ARN for a KMS key",
        )

    @staticmethod
    def __valid_iam_role_arn(key) -> Regex:
        return Regex(
            r"^arn:aws:iam::\d{12}:role/.*",
            error=f"{key} must contain a valid ARN for an IAM role",
        )

    @staticmethod
    def __valid_dbt_email_address(key) -> Regex:
        return Regex(
            r"^[\w.-]+@(businessandtrade.gov.uk|digital.trade.gov.uk)$",
            error=f"{key} must contain a valid DBT email address",
        )

    @staticmethod
    def __no_configuration_required_schema(schema_type) -> Schema:
        return Schema({"type": schema_type, Optional("services"): Or("__all__", [str])})


class ConditionalOpensSearchSchema(Schema):
    # TODO: DBTP-1943: Move to OpenSearch provider?
    _valid_opensearch_min_volume_size: int = 10

    # TODO: DBTP-1943: Move to OpenSearch provider?
    _valid_opensearch_max_volume_size: dict = {
        "tiny": 100,
        "small": 200,
        "small-ha": 200,
        "medium": 512,
        "medium-ha": 512,
        "large": 1000,
        "large-ha": 1000,
        "x-large": 1500,
        "x-large-ha": 1500,
    }

    def validate(self, data, _is_conditional_schema=True) -> Schema:
        data = super(ConditionalOpensSearchSchema, self).validate(
            data, _is_conditional_schema=False
        )
        if _is_conditional_schema:
            default_plan = None
            default_volume_size = None

            default_environment_config = data["environments"].get(
                "*", data["environments"].get("default", None)
            )
            if default_environment_config:
                default_plan = default_environment_config.get("plan", None)
                default_volume_size = default_environment_config.get("volume_size", None)

            for env in data["environments"]:
                volume_size = data["environments"][env].get("volume_size", default_volume_size)
                plan = data["environments"][env].get("plan", default_plan)

                if volume_size:
                    if not plan:
                        # TODO: DBTP-1943: Raise suitable PlatformException?
                        raise SchemaError(f"Missing key: 'plan'")

                    if volume_size < self._valid_opensearch_min_volume_size:
                        # TODO: DBTP-1943: Raise suitable PlatformException?
                        raise SchemaError(
                            f"Key 'environments' error: Key '{env}' error: Key 'volume_size' error: should be an integer greater than {self._valid_opensearch_min_volume_size}"
                        )

                    for key in self._valid_opensearch_max_volume_size:
                        if (
                            plan == key
                            and not volume_size <= self._valid_opensearch_max_volume_size[key]
                        ):
                            # TODO: DBTP-1943: Raise suitable PlatformException?
                            raise SchemaError(
                                f"Key 'environments' error: Key '{env}' error: Key 'volume_size' error: should be an integer between {self._valid_opensearch_min_volume_size} and {self._valid_opensearch_max_volume_size[key]} for plan {plan}"
                            )

        return data



import json
from datetime import datetime
from importlib.metadata import version
from pathlib import Path

from dbt_platform_helper.constants import EXTENSIONS_MODULE_PATH
from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.constants import SUPPORTED_AWS_PROVIDER_VERSION
from dbt_platform_helper.constants import SUPPORTED_TERRAFORM_VERSION
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.files import FileProvider
from dbt_platform_helper.providers.io import ClickIOProvider


class TerraformManifestProvider:
    def __init__(
        self, file_provider: FileProvider = FileProvider(), io: ClickIOProvider = ClickIOProvider()
    ):
        self.file_provider = file_provider
        self.io = io

    def generate_service_config(
        self,
        config_object,
        environment,
        platform_helper_version: str,
        platform_config,
        module_source_override: str = None,
    ):

        service_dir = f"terraform/services/{environment}/{config_object.name}"
        platform_config = ConfigProvider.apply_environment_defaults(platform_config)
        account = self._get_account_for_env(environment, platform_config)
        deploy_to_account_id = self._get_account_id_for_account(account, platform_config)
        application_name = platform_config["application"]

        terraform = {}
        self._add_header(terraform)

        self._add_service_locals(terraform, environment)

        self._add_provider(terraform, account, deploy_to_account_id)
        self._add_backend(
            terraform,
            platform_config,
            account,
            f"tfstate/application/{application_name}/services/{environment}/{config_object.name}.tfstate",
        )

        self._add_service_module(terraform, platform_helper_version, module_source_override)

        self._write_terraform_json(terraform, service_dir)

    def _add_service_locals(self, terraform, environment):
        terraform["locals"] = {
            "environment": environment,
            "platform_config": '${yamldecode(file("../../../../platform-config.yml"))}',
            "application": '${local.platform_config["application"]}',
            "environments": '${local.platform_config["environments"]}',
            "env_config": '${{for name, config in local.environments: name => merge(lookup(local.environments, "*", {}), config)}}',
            "service_config": '${yamldecode(file("./service-config.yml"))}',
            "raw_env_config": '${local.platform_config["environments"]}',
            "combined_env_config": '${{for name, config in local.raw_env_config: name => merge(lookup(local.raw_env_config, "*", {}), config)}}',
            "service_deployment_mode": '${lookup(local.combined_env_config[local.environment], "service-deployment-mode", "copilot")}',
            "non_copilot_service_deployment_mode": '${local.service_deployment_mode == "dual-deploy-copilot-traffic" || local.service_deployment_mode == "dual-deploy-platform-traffic" || local.service_deployment_mode == "platform" ? 1 : 0}',
            "custom_iam_policy_path": '${abspath(format("%s/../../../../services/%s/custom-iam-policy/%s.yml", path.module, local.service_config.name, local.environment))}',
            "custom_iam_policy_json": "${fileexists(local.custom_iam_policy_path) ? jsonencode(yamldecode(file(local.custom_iam_policy_path))) : null}",
        }

    def _add_service_module(
        self, terraform: dict, platform_helper_version: str, module_source_override: str = None
    ):
        source = (
            module_source_override
            or f"git::git@github.com:uktrade/platform-tools.git//terraform/ecs-service?depth=1&ref={platform_helper_version}"
        )
        terraform["module"] = {
            "ecs-service": {
                "source": source,
                "count": "${local.non_copilot_service_deployment_mode}",
                "application": "${local.application}",
                "environment": "${local.environment}",
                "service_config": "${local.service_config}",
                "env_config": "${local.env_config}",
                "platform_extensions": '${local.platform_config["extensions"]}',
                "custom_iam_policy_json": "${local.custom_iam_policy_json}",
            }
        }

    def generate_codebase_pipeline_config(
        self,
        platform_config: dict,
        platform_helper_version: str,
        ecr_imports: dict[str, str],
        deploy_repository: str,
        module_source: str,
        workspace: str = None,
    ):
        default_account = self._get_account_for_env("*", platform_config)
        deploy_to_account_id = self._get_account_id_for_account(default_account, platform_config)
        state_key_suffix = f"{platform_config['application']}-codebase-pipelines"

        terraform = {}
        self._add_header(terraform)

        platform_config_file_name = (
            f"platform-config.{workspace}.yml" if workspace else PLATFORM_CONFIG_FILE
        )
        self._add_codebase_pipeline_locals(terraform, platform_config_file_name)
        self._add_provider(terraform, default_account, deploy_to_account_id)
        self._add_backend(
            terraform,
            platform_config,
            default_account,
            f"tfstate/application/{state_key_suffix}.tfstate",
        )
        self._add_codebase_pipeline_module(
            terraform, platform_helper_version, deploy_repository, module_source
        )
        self._add_imports(terraform, ecr_imports)
        self._add_checks(terraform, workspace)
        self._write_terraform_json(terraform, "terraform/codebase-pipelines")

    def generate_environment_config(
        self,
        platform_config: dict,
        env: str,
        platform_helper_version: str,
        module_source_override: str = None,
    ):
        platform_config = ConfigProvider.apply_environment_defaults(platform_config)
        account = self._get_account_for_env(env, platform_config)

        application_name = platform_config["application"]
        state_key_suffix = f"{platform_config['application']}-{env}"
        env_dir = f"terraform/environments/{env}"

        terraform = {}
        self._add_header(terraform)
        self._add_environment_locals(terraform, application_name)
        self._add_backend(
            terraform, platform_config, account, f"tfstate/application/{state_key_suffix}.tfstate"
        )
        self._add_extensions_module(terraform, platform_helper_version, env, module_source_override)
        self._add_moved(terraform, platform_config)
        self._ensure_no_hcl_manifest_file(env_dir)
        self._write_terraform_json(terraform, env_dir)

    @staticmethod
    def _get_account_for_env(env, platform_config):
        account = (
            platform_config.get("environments", {})
            .get(env, {})
            .get("accounts", {})
            .get("deploy", {})
            .get("name")
        )
        return account

    @staticmethod
    def _get_account_id_for_account(account_name, platform_config):
        environment_config = platform_config["environments"]
        account_id_lookup = {
            env["accounts"]["deploy"]["name"]: env["accounts"]["deploy"]["id"]
            for env in environment_config.values()
            if env is not None and "accounts" in env and "deploy" in env["accounts"]
        }
        return account_id_lookup.get(account_name)

    @staticmethod
    def _add_header(terraform: dict):
        time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        version_header = f"Generated by platform-helper {version('dbt-platform-helper')} / {time}."
        warning = "WARNING: This is an autogenerated file, not for manual editing."
        # The "//" key denotes a comment in terraform json.
        terraform["//"] = f"{version_header} {warning}"

    @staticmethod
    def _add_codebase_pipeline_locals(terraform: dict, file_name: str):
        decode_statement = f'${{yamldecode(file("../../{file_name}"))}}'
        terraform["locals"] = {
            "platform_config": decode_statement,
            "application": '${local.platform_config["application"]}',
            "all_codebases": '${local.platform_config["codebase_pipelines"]}',
            "environments": '${local.platform_config["environments"]}',
        }

    @staticmethod
    def _add_provider(terraform: dict, deploy_to_account: str, deploy_to_account_id: str):
        terraform["provider"] = {"aws": {}}
        terraform["provider"]["aws"]["region"] = "eu-west-2"
        terraform["provider"]["aws"]["profile"] = deploy_to_account
        terraform["provider"]["aws"]["allowed_account_ids"] = [deploy_to_account_id]

    @staticmethod
    def _add_backend(terraform: dict, platform_config: dict, account: str, state_key: str):
        terraform["terraform"] = {
            "required_version": SUPPORTED_TERRAFORM_VERSION,
            "backend": {
                "s3": {
                    "bucket": f"terraform-platform-state-{account}",
                    "key": state_key,
                    "region": "eu-west-2",
                    "encrypt": True,
                    "kms_key_id": f"alias/terraform-platform-state-s3-key-{account}",
                    "dynamodb_table": f"terraform-platform-lockdb-{account}",
                }
            },
            "required_providers": {
                "aws": {"source": "hashicorp/aws", "version": SUPPORTED_AWS_PROVIDER_VERSION}
            },
        }

    @staticmethod
    def _add_codebase_pipeline_module(
        terraform: dict,
        platform_helper_version: str,
        deploy_repository: str,
        module_source: str,
    ):
        source = module_source
        terraform["module"] = {
            "codebase-pipelines": {
                "source": source,
                "for_each": "${local.all_codebases}",
                "application": "${local.application}",
                "codebase": "${each.key}",
                "repository": "${each.value.repository}",
                "deploy_repository": f"{deploy_repository}",
                "deploy_repository_branch": '${lookup(each.value, "deploy_repository_branch", "main")}',
                "additional_ecr_repository": '${lookup(each.value, "additional_ecr_repository", null)}',
                "cache_invalidation": '${lookup(each.value, "cache_invalidation", null)}',
                "pipelines": '${lookup(each.value, "pipelines", [])}',
                "services": "${each.value.services}",
                "requires_image_build": '${lookup(each.value, "requires_image_build", true)}',
                "slack_channel": '${lookup(each.value, "slack_channel", "/codebuild/slack_oauth_channel")}',
                "env_config": "${local.environments}",
                "platform_tools_version": f"{platform_helper_version}",
            }
        }

    @staticmethod
    def _add_extensions_module(
        terraform: dict, platform_helper_version: str, env: str, module_source_override: str = None
    ):
        source = module_source_override or f"{EXTENSIONS_MODULE_PATH}{platform_helper_version}"
        terraform["module"] = {
            "extensions": {
                "source": source,
                "args": "${local.args}",
                "environment": env,
                "repos": "${concat(local.codebase_pipeline_repos != null ? (distinct(values(local.codebase_pipeline_repos))) : null, try([local.config.deploy_repository], []))}",
            }
        }

    @staticmethod
    def _add_imports(terraform: dict, ecr_imports: dict[str, str]):
        if ecr_imports:
            terraform["import"] = {
                "for_each": "${%s}" % json.dumps(ecr_imports),
                "id": "${each.value}",
                "to": "module.codebase-pipelines[each.key].aws_ecr_repository.this",
            }

    @staticmethod
    def _add_checks(terraform: dict, workspace: str):
        if workspace:
            terraform["resource"] = {
                "terraform_data": {
                    "workspace_check": {
                        "lifecycle": {
                            "precondition": {
                                "condition": f'${{terraform.workspace == "{workspace}"}}',
                                "error_message": f"Must be in {workspace} workspace",
                            }
                        }
                    }
                }
            }

    @staticmethod
    def _add_environment_locals(terraform: dict, app: str):
        terraform["locals"] = {
            "config": '${yamldecode(file("../../../platform-config.yml"))}',
            "environments": '${local.config["environments"]}',
            "env_config": '${{for name, config in local.environments: name => merge(lookup(local.environments, "*", {}), config)}}',
            "args": {
                "application": app,
                "services": '${local.config["extensions"]}',
                "env_config": "${local.env_config}",
            },
            "codebase_pipeline_repos": '${try({for k, v in local.config["codebase_pipelines"]: k => v.repository}, null)}',
        }

    @staticmethod
    def _add_moved(terraform, platform_config):
        extensions_comment = "Moved extensions-tf to just extensions - this block tells terraform this. Can be removed once all services have moved to the new naming."
        terraform["moved"] = [
            {
                "//": extensions_comment,
                "from": "module.extensions-tf",
                "to": "module.extensions",
            }
        ]

        extensions = platform_config.get("extensions", {})
        s3_extension_names = [
            extension_name
            for extension_name, extension in extensions.items()
            if extension["type"] == "s3"
        ]
        s3_comment = "S3 bucket resources are now indexed. Can be removed once all services have moved to terraform-platform-modules 5.x."

        for name in s3_extension_names:
            resources = [
                "aws_s3_bucket_server_side_encryption_configuration.encryption-config",
                "aws_s3_bucket_policy.bucket-policy",
                "aws_kms_key.kms-key",
                "aws_kms_alias.s3-bucket",
            ]
            moves = [f'module.extensions.module.s3["{name}"].{resource}' for resource in resources]
            for move in moves:
                terraform["moved"].append(
                    {
                        "//": s3_comment,
                        "from": move,
                        "to": f"{move}[0]",
                    }
                )

    def _write_terraform_json(self, terraform: dict, env_dir: str):
        message = self.file_provider.mkfile(
            str(Path(env_dir)),
            "main.tf.json",
            json.dumps(terraform, indent=2),
            True,
        )
        self.io.info(message)

    def _ensure_no_hcl_manifest_file(self, env_dir):
        message = self.file_provider.delete_file(env_dir, "main.tf")
        if message:
            self.io.info(f"Manifest has moved to main.tf.json. {message}")



from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from slack_sdk.models import blocks

from dbt_platform_helper.platform_exception import PlatformException


class SlackChannelNotifierException(PlatformException):
    pass


class SlackChannelNotifier:
    def __init__(self, slack_token: str, slack_channel_id: str):
        self.client = WebClient(slack_token)
        self.slack_channel_id = slack_channel_id

    def post_update(self, message_ref, message, context=None):
        args = {
            "channel": self.slack_channel_id,
            "blocks": self._build_message_blocks(message, context),
            "text": message,
            "unfurl_links": False,
            "unfurl_media": False,
        }

        try:
            response = self.client.chat_update(ts=message_ref, **args)
            return response["ts"]
        except SlackApiError as e:
            raise SlackChannelNotifierException(f"Slack notification unsuccessful: {e}")

    def post_new(self, message, context=None, title=None, reply_broadcast=None, thread_ref=None):
        args = {
            "channel": self.slack_channel_id,
            "blocks": self._build_message_blocks(message, context),
            "text": title if title else message,
            "reply_broadcast": reply_broadcast,
            "unfurl_links": False,
            "unfurl_media": False,
            "thread_ts": thread_ref,
        }

        try:
            response = self.client.chat_postMessage(ts=None, **args)
            return response["ts"]
        except SlackApiError as e:
            raise SlackChannelNotifierException(f"Slack notification unsuccessful: {e}")

    def _build_message_blocks(self, message, context):
        message_blocks = [
            blocks.SectionBlock(
                text=blocks.TextObject(type="mrkdwn", text=message),
            ),
        ]

        if context:
            message_blocks.append(
                blocks.ContextBlock(
                    elements=[blocks.TextObject(type="mrkdwn", text=element) for element in context]
                )
            )
        return message_blocks



import boto3


class Opensearch:

    def __init__(self, client: boto3.client):
        self.client = client
        self.engine = "OpenSearch"

    def get_reference(self) -> str:
        return self.engine.lower()

    def get_supported_versions(self) -> list[str]:
        response = self.client.list_versions()
        all_versions = response["Versions"]

        supported_versions = [
            version.removeprefix(f"{self.engine}_")
            for version in all_versions
            if version.startswith(f"{self.engine}_")
        ]

        return supported_versions



from collections import defaultdict

import botocore
from boto3 import Session

from dbt_platform_helper.providers.aws.exceptions import AWSException
from dbt_platform_helper.providers.aws.exceptions import ImageNotFoundException
from dbt_platform_helper.providers.aws.exceptions import MultipleImagesFoundException
from dbt_platform_helper.providers.aws.exceptions import RepositoryNotFoundException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.utils.aws import get_aws_session_or_abort

NOT_A_UNIQUE_TAG_INFO = 'INFO: The tag "{image_ref}" is not a unique, commit-specific tag. Deploying the corresponding commit tag "{commit_tag}" instead.'
NO_ASSOCIATED_COMMIT_TAG_WARNING = 'WARNING: The AWS ECR image "{image_ref}" has no associated commit tag so deploying "{image_ref}". Note this could result in images with unintended or incompatible changes being deployed in new ECS Tasks for your service.'


class ECRProvider:
    def __init__(self, session: Session = None, click_io: ClickIOProvider = ClickIOProvider()):
        self.session = session
        self.click_io = click_io

    def get_ecr_repo_names(self) -> list[str]:
        out = []
        for page in self._get_client().get_paginator("describe_repositories").paginate():
            out.extend([repo["repositoryName"] for repo in page.get("repositories", {})])
        return out

    def get_commit_tag_for_reference(self, application_name: str, codebase: str, image_ref: str):
        repository = f"{application_name}/{codebase}"
        next_page_token = None
        tag_map = {}
        digest_map = defaultdict(dict)

        while True:
            image_list = self._get_ecr_images(repository, image_ref, next_page_token)
            next_page_token = image_list.get("nextToken")

            for image in image_list["imageIds"]:
                digest, tag = image["imageDigest"], image["imageTag"]
                digest_map[digest][tag.split("-")[0]] = tag
                tag_map[tag] = digest

            if not next_page_token:
                break

        if image_ref.startswith("commit-"):
            if image_ref in tag_map:
                return image_ref
            else:
                candidates = [
                    tag
                    for tag in tag_map.keys()
                    if image_ref.startswith(tag) or tag.startswith(image_ref)
                ]
                if not candidates:
                    raise ImageNotFoundException(image_ref)
                if len(candidates) > 1:
                    raise MultipleImagesFoundException(image_ref, candidates)
                return candidates[0]
        else:
            digest = tag_map.get(image_ref)
            if not digest:
                raise ImageNotFoundException(image_ref)

            commit_tag = digest_map.get(digest, dict()).get("commit")

            if commit_tag:
                self.click_io.info(
                    NOT_A_UNIQUE_TAG_INFO.format(image_ref=image_ref, commit_tag=commit_tag)
                )
                return commit_tag
            else:
                self.click_io.warn(NO_ASSOCIATED_COMMIT_TAG_WARNING.format(image_ref=image_ref))
                return image_ref

    def _get_ecr_images(self, repository, image_ref, next_page_token):
        params = {"repositoryName": repository, "filter": {"tagStatus": "TAGGED"}}
        if next_page_token:
            params["nextToken"] = next_page_token
        try:
            image_list = self._get_client().list_images(**params)
            return image_list
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "RepositoryNotFoundException":
                raise RepositoryNotFoundException(repository)
            else:
                raise AWSException(
                    f"Unexpected error for repo '{repository}' and image reference '{image_ref}': {e}"
                )

    @staticmethod
    def _check_image_details_exists(image_info: dict, image_ref: str):
        """Error handling for any unexpected scenario where AWS ECR returns a
        malformed response."""

        if "imageDetails" not in image_info:
            raise ImageNotFoundException(image_ref)

    def _get_client(self):
        if not self.session:
            self.session = get_aws_session_or_abort()
        return self.session.client("ecr")



# TODO: DBTP-1888: Move to Config provider
REPO_ORG = "uktrade"
PLATFORM_CONFIG_FILE = "platform-config.yml"
SERVICE_DIRECTORY = "services"
SERVICE_CONFIG_FILE = "service-config.yml"
ENVIRONMENT_PIPELINE_MODULE_PATH = (
    "git::git@github.com:uktrade/platform-tools.git//terraform/environment-pipelines?depth=1&ref="
)

CODEBASE_PIPELINE_MODULE_PATH = (
    "git::git@github.com:uktrade/platform-tools.git//terraform/codebase-pipelines?depth=1&ref="
)
EXTENSIONS_MODULE_PATH = (
    "git::git@github.com:uktrade/platform-tools.git//terraform/extensions?depth=1&ref="
)
PLATFORM_HELPER_VERSION_OVERRIDE_KEY = "PLATFORM_HELPER_VERSION_OVERRIDE"
TERRAFORM_EXTENSIONS_MODULE_SOURCE_OVERRIDE_ENV_VAR = "TERRAFORM_EXTENSIONS_MODULE_SOURCE_OVERRIDE"
TERRAFORM_ENVIRONMENT_PIPELINES_MODULE_SOURCE_OVERRIDE_ENV_VAR = (
    "TERRAFORM_ENVIRONMENT_PIPELINES_MODULE_SOURCE_OVERRIDE"
)
TERRAFORM_CODEBASE_PIPELINES_MODULE_SOURCE_OVERRIDE_ENV_VAR = (
    "TERRAFORM_CODEBASE_PIPELINES_MODULE_SOURCE_OVERRIDE"
)
TERRAFORM_ECS_SERVICE_MODULE_SOURCE_OVERRIDE_ENV_VAR = (
    "TERRAFORM_ECS_SERVICE_MODULE_SOURCE_OVERRIDE"
)
TERRAFORM_MODULE_SOURCE_TYPE_ENV_VAR = "TERRAFORM_MODULE_SOURCE_TYPE"  # "LOCAL", "SSH", "OVERRIDE"
IMAGE_TAG_ENV_VAR = "IMAGE_TAG"
PLATFORM_HELPER_PACKAGE_NAME = "dbt-platform-helper"
SUPPORTED_TERRAFORM_VERSION = "~> 1.11"
SUPPORTED_AWS_PROVIDER_VERSION = "~> 6"
FIRST_UPGRADABLE_PLATFORM_HELPER_MAJOR_VERSION = 13

MERGED_TPM_PLATFORM_HELPER_VERSION = 14
PLATFORM_CONFIG_SCHEMA_VERSION = 1
SERVICE_CONFIG_SCHEMA_VERSION = 1

# Keys
CODEBASE_PIPELINES_KEY = "codebase_pipelines"
ENVIRONMENTS_KEY = "environments"
ENVIRONMENT_PIPELINES_KEY = "environment_pipelines"

# Conduit
CONDUIT_ADDON_TYPES = [
    "opensearch",
    "postgres",
    "redis",
]
CONDUIT_DOCKER_IMAGE_LOCATION = "public.ecr.aws/uktrade/tunnel"
HYPHENATED_APPLICATION_NAME = "hyphenated-application-name"
ALPHANUMERIC_ENVIRONMENT_NAME = "alphanumericenvironmentname123"
ALPHANUMERIC_SERVICE_NAME = "alphanumericservicename123"
COPILOT_IDENTIFIER = "c0PIlotiD3ntIF3r"
CLUSTER_NAME_SUFFIX = f"Cluster-{COPILOT_IDENTIFIER}"
SERVICE_NAME_SUFFIX = f"Service-{COPILOT_IDENTIFIER}"
REFRESH_TOKEN_MESSAGE = (
    "To refresh this SSO session run `aws sso login` with the corresponding profile"
)
COPILOT_RULE_PRIORITY = 48000
PLATFORM_RULE_STARTING_PRIORITY = 10000
RULE_PRIORITY_INCREMENT = 100
DUMMY_RULE_REASON = "DummyRule"
MAINTENANCE_PAGE_TAGS = ["MaintenancePage", "AllowedIps", "BypassIpFilter", "AllowedSourceIps"]
MAINTENANCE_PAGE_REASON = "MaintenancePage"
MANAGED_BY_PLATFORM = "DBT Platform"
MANAGED_BY_SERVICE_TERRAFORM = "DBT Platform - Service Terraform"
MANAGED_BY_PLATFORM_TERRAFORM = "DBT Platform - Terraform"
STANDARD_PLATFORM_SSO_ROLES = [
    "AdministratorAccess",
    "DBTPlatformDeveloperWrite",
    "DBTPlatformDeveloperRead",
]
ROUTED_TO_PLATFORM_MODES = ["platform", "dual-deploy-platform-traffic"]
ROUTED_TO_COPILOT_MODES = ["copilot", "dual-deploy-copilot-traffic"]
HTTP_SERVICE_TYPES = ["Load Balanced Web Service"]



from pathlib import Path

from dbt_platform_helper.providers.yaml_file import YamlFileProvider


class PlanLoader:

    PROJECT_DIR = Path(__file__).resolve().parent.parent.parent

    def __init__(
        self,
        extensions: dict = None,
        terraform_dir: str = "terraform",
        loader: YamlFileProvider = YamlFileProvider,
    ):
        self.path = terraform_dir
        self.loader = loader
        self._cache = {}
        self.extensions = extensions or {
            "redis": "elasticache-redis",
            "opensearch": "opensearch",
            "postgres": "postgres",
        }

    def load(self):
        result = {}
        for key, value in self.extensions.items():
            result[key] = self._load_plan(key, f"{self.PROJECT_DIR}/{self.path}/{value}/plans.yml")
        return result

    def _load_plan(self, name, path):
        if name in self._cache:
            return self._cache[name]
        else:
            plan = self.loader.load(path)
            self._cache[name] = plan
            return plan

    def get_plan_names(self, extension):
        plans = self.load()
        return list(plans[extension].keys())



from collections import defaultdict
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import List

from botocore.exceptions import ClientError

from dbt_platform_helper.constants import COPILOT_RULE_PRIORITY
from dbt_platform_helper.constants import DUMMY_RULE_REASON
from dbt_platform_helper.constants import HTTP_SERVICE_TYPES
from dbt_platform_helper.constants import MAINTENANCE_PAGE_REASON
from dbt_platform_helper.constants import MANAGED_BY_PLATFORM
from dbt_platform_helper.constants import MANAGED_BY_SERVICE_TERRAFORM
from dbt_platform_helper.constants import PLATFORM_RULE_STARTING_PRIORITY
from dbt_platform_helper.constants import RULE_PRIORITY_INCREMENT
from dbt_platform_helper.domain.service import ServiceManager
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.config_validator import ConfigValidator
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.load_balancers import LoadBalancerProvider
from dbt_platform_helper.utils.application import load_application


class RollbackException(PlatformException):
    pass


class RuleType(Enum):
    PLATFORM = "platform"
    MAINTENANCE = "maintenance"
    DEFAULT = "default"
    COPILOT = "copilot"
    MANUAL = "manual"
    DUMMY = "dummy"


class Deployment(Enum):
    PLATFORM = "platform"
    COPILOT = "copilot"
    DUAL_DEPLOY_PLATFORM = "dual-deploy-platform-traffic"
    DUAL_DEPLOY_COPILOT = "dual-deploy-copilot-traffic"


@dataclass
class OperationState:
    created_rules: List[object] = field(default_factory=list)
    deleted_rules: List[object] = field(default_factory=list)
    updated_rules: List[object] = field(default_factory=list)
    listener_arn: str = ""


class UpdateALBRules:

    def __init__(
        self,
        session,
        config_provider: ConfigProvider = ConfigProvider(ConfigValidator()),
        io: ClickIOProvider = ClickIOProvider(),
        load_application=load_application,
        load_balancer_p: LoadBalancerProvider = LoadBalancerProvider,
    ):
        self.config_provider = config_provider
        self.io = io
        self.load_application = load_application
        self.load_balancer: LoadBalancerProvider = load_balancer_p(session, io=self.io)

    def update_alb_rules(
        self,
        environment: str,
    ):
        """
        Change ALB rules for a given environment.

        Attempt to rollback the rules created/deleted if a failure occurs.
        """

        operation_state = OperationState()

        try:
            self._execute_rule_updates(environment, operation_state)
            if operation_state.created_rules:
                self.io.info(f"Created rules: {len(operation_state.created_rules)}")
                self._output_rule_changes(operation_state.created_rules)
            if operation_state.deleted_rules:
                self.io.info(f"Deleted rules: {len(operation_state.deleted_rules)}")
                self._output_rule_changes(operation_state.deleted_rules)
            if not operation_state.created_rules and not operation_state.deleted_rules:
                self.io.info("No rule updates required")
        except Exception as e:
            if operation_state.created_rules or operation_state.deleted_rules:
                self.io.error(f"Error during rule update: {str(e)}")
                self.io.info("Attempting to rollback changes ...")
                try:
                    self._rollback_changes(operation_state)
                    raise PlatformException(f"Rule update failed and rolled back")
                except RollbackException as rollback_error:
                    raise PlatformException(f"Rollback failed: \n{str(rollback_error)}")
            else:
                raise

    def _execute_rule_updates(self, environment: str, operation_state: OperationState):
        platform_config = self.config_provider.get_enriched_config()

        application_name = platform_config.get("application", "")
        application = self.load_application(app=application_name)

        service_deployment_mode = (
            platform_config.get("environments")
            .get(environment, {})
            .get("service-deployment-mode", Deployment.COPILOT.value)
        )

        self.io.info(f"Deployment Mode: {service_deployment_mode}")

        listener_arn = self.load_balancer.get_https_listener_for_application(
            application.name, environment
        )

        operation_state.listener_arn = listener_arn

        self.io.debug(f"Listener ARN: {listener_arn}")

        rules = self.load_balancer.get_rules_with_tags_by_listener_arn(listener_arn)

        mapped_rules = {
            rule_type: [rule for rule in rules if self._filter_rule_type(rule) == rule_type]
            for rule_type in set(self._filter_rule_type(rule) for rule in rules)
        }

        if mapped_rules.get(RuleType.MANUAL.value, []):
            rule_arns = [rule["RuleArn"] for rule in mapped_rules[RuleType.MANUAL.value]]
            message = f"""The following rules have been created manually please review and if required set 
            the rules priority to the copilot range after priority: {COPILOT_RULE_PRIORITY}.\n
            Rules: {rule_arns}"""
            raise PlatformException(message)

        if (
            service_deployment_mode == Deployment.PLATFORM.value
            or service_deployment_mode == Deployment.DUAL_DEPLOY_PLATFORM.value
        ):
            service_models = ServiceManager().get_service_models(application, environment)
            grouped = defaultdict(list)

            service_mapped_tgs = self._get_tg_arns_for_platform_services(
                application_name, environment
            )

            for service in service_models:
                if service.type not in HTTP_SERVICE_TYPES or service.name not in service_mapped_tgs:
                    continue

                additional_rules = getattr(service.http, "additional_rules", None)
                if additional_rules:
                    for rule in additional_rules:
                        grouped[(service.name, rule.path)].extend(rule.alias)

                grouped[(service.name, service.http.path)].extend(service.http.alias)

            rules = []
            for (name, path), aliases in grouped.items():
                path_pattern = ["/*"] if path == "/" else [path, f"{path}/*"]
                condition_length = len(aliases) + len(path_pattern)

                # AWS allows a maximum of 5 condition values per rule, this includes host and path values
                max_conditions = 5
                if condition_length > max_conditions:
                    i = 0
                    while i < condition_length:
                        remaining_slots = max_conditions - len(path_pattern)
                        alias_split = aliases[i : i + remaining_slots] if aliases else []
                        if not alias_split:
                            break

                        rules.append(
                            {
                                "service": name,
                                "path": path,
                                "path_pattern": path_pattern,
                                "aliases": sorted(set(alias_split)),
                            }
                        )
                        i += remaining_slots
                else:
                    rules.append(
                        {
                            "service": name,
                            "path": path,
                            "path_pattern": path_pattern,
                            "aliases": sorted(set(aliases)),
                        }
                    )

            rules.sort(
                key=lambda r: (len([s for s in r["path"].split("/") if s]), r["aliases"]),
                reverse=True,
            )

            platform_rules = mapped_rules.get(RuleType.PLATFORM.value, [])

            rule_priority = PLATFORM_RULE_STARTING_PRIORITY
            for rule in rules:
                actions = [
                    {"Type": "forward", "TargetGroupArn": service_mapped_tgs[rule["service"]]}
                ]
                conditions = [
                    {"Field": "host-header", "HostHeaderConfig": {"Values": rule["aliases"]}},
                    {
                        "Field": "path-pattern",
                        "PathPatternConfig": {"Values": rule["path_pattern"]},
                    },
                ]
                tags = [
                    {"Key": "application", "Value": application_name},
                    {"Key": "environment", "Value": environment},
                    {"Key": "service", "Value": rule["service"]},
                    {"Key": "reason", "Value": "service"},
                    {"Key": "managed-by", "Value": MANAGED_BY_PLATFORM},
                ]

                try:
                    rule = self.load_balancer.create_rule(
                        listener_arn,
                        actions,
                        conditions,
                        rule_priority,
                        tags,
                    )["Rules"][0]
                    rule_arn = rule["RuleArn"]

                    if rule_arn in [rule["RuleArn"] for rule in platform_rules]:
                        operation_state.updated_rules.append(rule)
                    else:
                        operation_state.created_rules.append(rule)

                except ClientError as e:
                    if e.response["Error"]["Code"] == "PriorityInUse":
                        existing_rule = [
                            r for r in platform_rules if r["Priority"] == str(rule_priority)
                        ]
                        if not existing_rule:
                            raise PlatformException(
                                f"Priority {rule_priority} in use but no matching platform rule found."
                            )

                        self._delete_rules(existing_rule, operation_state)

                        rule = self.load_balancer.create_rule(
                            listener_arn,
                            actions,
                            conditions,
                            rule_priority,
                            tags,
                        )["Rules"][0]

                        operation_state.created_rules.append(rule)
                    else:
                        raise

                rule_priority += RULE_PRIORITY_INCREMENT

            # Remove dangling rules
            managed_rules = [
                rule["RuleArn"]
                for rules in [
                    operation_state.created_rules,
                    operation_state.updated_rules,
                    operation_state.deleted_rules,
                ]
                for rule in rules
            ]

            for rule in platform_rules:
                if rule["RuleArn"] not in managed_rules:
                    self._delete_rules([rule], operation_state)

        # Remove dummy rules
        if service_deployment_mode == Deployment.PLATFORM.value:
            self._delete_rules(mapped_rules.get(RuleType.DUMMY.value, []), operation_state)

        if (
            service_deployment_mode == Deployment.COPILOT.value
            or service_deployment_mode == Deployment.DUAL_DEPLOY_COPILOT.value
        ):
            rules_to_delete = mapped_rules.get(RuleType.PLATFORM.value, [])
            if not rules_to_delete:
                return

            self.io.warn("Platform rules will be deleted")
            self._output_rule_changes(rules_to_delete)
            if self.io.confirm(
                f"This command is destructive and will remove load balancer listener rules created by the platform, you may lose access to your services. Are you sure you want to continue?"
            ):
                self._delete_rules(rules_to_delete, operation_state)

    def _delete_rules(self, rules: List[dict], operation_state: OperationState):
        for rule in rules:
            rule_arn = rule["RuleArn"]
            try:
                self.load_balancer.delete_listener_rule_by_resource_arn(rule_arn)
                operation_state.deleted_rules.append(rule)
                self.io.debug(f"Deleted existing rule: {rule_arn}")
            except Exception as e:
                self.io.error(f"Failed to delete existing rule {rule_arn}: {str(e)}")
                raise

    def _filter_rule_type(self, rule: dict) -> str:
        if rule["Tags"]:
            if rule["Tags"].get("managed-by", "") == MANAGED_BY_PLATFORM:
                return RuleType.PLATFORM.value
            if rule["Tags"].get("reason", None) == MAINTENANCE_PAGE_REASON:
                return RuleType.MAINTENANCE.value
            if rule["Tags"].get("reason", None) == DUMMY_RULE_REASON:
                return RuleType.DUMMY.value

        if rule["Priority"] == "default":
            return RuleType.DEFAULT.value
        if int(rule["Priority"]) >= COPILOT_RULE_PRIORITY:
            return RuleType.COPILOT.value

        return RuleType.MANUAL.value

    def _get_tg_arns_for_platform_services(
        self, application: str, environment: str, managed_by: str = MANAGED_BY_SERVICE_TERRAFORM
    ) -> dict:
        tgs = self.load_balancer.get_target_groups_with_tags([])
        service_mapped_tgs = {}
        for tg in tgs:
            if (
                tg["Tags"].get("application", "") == application
                and tg["Tags"].get("environment", "") == environment
                and tg["Tags"].get("managed-by", "") == managed_by
            ):
                if tg["Tags"].get("service", ""):
                    service_mapped_tgs[tg["Tags"].get("service")] = tg["TargetGroupArn"]
                else:
                    tg_name = tg["TargetGroupName"]
                    self.io.warn(f"Target group {tg_name} has no 'service' tag")
        return service_mapped_tgs

    def _rollback_changes(self, operation_state: OperationState):
        rollback_errors = []
        delete_rollbacks = []
        create_rollbacks = []
        for rule in operation_state.created_rules:
            rule_arn = rule["RuleArn"]
            try:
                self.io.debug(f"Rolling back: Deleting created rule {rule_arn}")
                self.load_balancer.delete_listener_rule_by_resource_arn(rule_arn)
                delete_rollbacks.append(rule_arn)
            except Exception as e:
                error_msg = f"Failed to delete rule {rule_arn} during rollback: {str(e)}"
                rollback_errors.append(error_msg)

        for rule_snapshot in operation_state.deleted_rules:
            rule_arn = rule_snapshot["RuleArn"]
            try:
                self.io.debug(f"Rolling back: Recreating deleted rule {rule_arn}")
                create_rollbacks.append(
                    self.load_balancer.create_rule(
                        operation_state.listener_arn,
                        actions=rule_snapshot["Actions"],
                        conditions=[
                            {"Field": key, "Values": value}
                            for key, value in rule_snapshot["Conditions"].items()
                        ],
                        priority=int(rule_snapshot["Priority"]),
                        tags=[
                            {"Key": key, "Value": value}
                            for key, value in rule_snapshot["Tags"].items()
                        ],
                    )["Rules"][0]["RuleArn"]
                )
            except Exception as e:
                error_msg = f"Failed to recreate rule {rule_arn} during rollback: {str(e)}"
                rollback_errors.append(error_msg)

        if rollback_errors:
            self.io.warn("Some rollback operations failed. Manual intervention may be required.")
            errors = "\n".join(rollback_errors)
            raise RollbackException(f"Rollback partially failed: {errors}")
        else:
            self.io.info("Rollback completed successfully")
            self.io.info(
                f"Rolled back rules by creating: {create_rollbacks} \n and deleting {delete_rollbacks}"
            )

    def _output_rule_changes(self, rules):
        for rule in rules:
            hosts = []
            paths = []

            conditions = rule.get("Conditions", [])
            if isinstance(conditions, list):
                for condition in conditions:
                    if condition["Field"] == "host-header":
                        hosts.extend(condition["HostHeaderConfig"]["Values"])
                    elif condition["Field"] == "path-pattern":
                        paths.extend(condition["PathPatternConfig"]["Values"])
            elif isinstance(conditions, dict):
                hosts.extend(conditions.get("host-header", ""))
                paths.extend(conditions.get("path-pattern", ""))

            rule_arn = rule["RuleArn"]
            priority = rule["Priority"]
            hosts = ",".join(hosts)
            paths = ",".join(paths)
            self.io.info(f"ARN: {rule_arn}")
            self.io.info(f"Priority: {priority}")
            self.io.info(f"Hosts: {hosts}")
            self.io.info(f"Paths: {paths}\n")



import click

from dbt_platform_helper.domain.service import ServiceManager
from dbt_platform_helper.domain.update_alb_rules import UpdateALBRules
from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.autoscaling import AutoscalingProvider
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.config_validator import ConfigValidator
from dbt_platform_helper.providers.ecs import ECS
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.logs import LogsProvider
from dbt_platform_helper.providers.s3 import S3Provider
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.click import ClickDocOptGroup


@click.group(cls=ClickDocOptGroup)
def internal():
    """Internal commands for use within pipelines or by Platform Team."""


@internal.command()
def migrate_service_manifests():
    """Migrate copilot manifests to service manifests."""
    click_io = ClickIOProvider()

    try:
        service_manager = ServiceManager()
        service_manager.migrate_copilot_manifests()
    except PlatformException as error:
        click_io.abort_with_error(str(error))


@internal.group(cls=ClickDocOptGroup)
def service():
    """Subgroup for 'internal service' commands."""


@service.command(help="Trigger an ECS deployment.")
@click.option("--name", required=True, help="The name of the ECS service to create or update.")
@click.option(
    "--env",
    required=True,
    help="The name of the environment where the ECS service will be created or updated.",
)
@click.option(
    "--image-tag",
    required=True,
    help="Image tag to deploy for the service(s). Takes precedence over the $IMAGE_TAG environment variable.",
)
def deploy(name, env, image_tag):
    """Register a new ECS task definition from an S3 JSON template, update the
    ECS service, and tail CloudWatch logs until the ECS rollout is complete."""
    click_io = ClickIOProvider()

    try:

        config = ConfigProvider(ConfigValidator()).get_enriched_config()
        application_name = config.get("application", "")
        application = load_application(app=application_name, env=env)

        ecs_client = application.environments[env].session.client("ecs")
        ssm_client = application.environments[env].session.client("ssm")
        s3_client = application.environments[env].session.client("s3")
        logs_client = application.environments[env].session.client("logs")
        autoscaling_client = application.environments[env].session.client("application-autoscaling")

        ecs_provider = ECS(
            ecs_client=ecs_client,
            ssm_client=ssm_client,
            application_name=application.name,
            env=env,
        )
        s3_provider = S3Provider(client=s3_client)
        logs_provider = LogsProvider(client=logs_client)
        autoscaling_provider = AutoscalingProvider(client=autoscaling_client)

        service_manager = ServiceManager(
            ecs_provider=ecs_provider,
            s3_provider=s3_provider,
            logs_provider=logs_provider,
            autoscaling_provider=autoscaling_provider,
        )
        service_manager.deploy(
            service=name,
            environment=env,
            application=application.name,
            image_tag=image_tag,
        )
    except PlatformException as error:
        click_io.abort_with_error(str(error))


@service.command(help="Generate Terraform manifest for the specified service(s).")
@click.option(
    "--name",
    required=False,
    help="The name of the service(s) to generate service manifest(s) for.",
    multiple=True,
)
@click.option(
    "--env",
    required=True,
    help="The name of the environment to generate service manifests for.",
)
def generate(name, env):
    """Validates the service-config.yml format, applies the environment-specific
    overrides, and generates a Terraform manifest at
    /terraform/services/<environment>/<service>/main.tf.json."""

    services = list(name)
    click_io = ClickIOProvider()

    try:
        service_manager = ServiceManager()
        service_manager.generate(environment=env, services=services)

    except PlatformException as err:
        click_io.abort_with_error(str(err))


@internal.group(cls=ClickDocOptGroup)
def alb():
    """Load Balancer related commands."""
    PlatformHelperVersioning().check_if_needs_update()


@alb.command()
@click.option("--env", type=str, required=True)
def update_rules(env: str):
    """Update alb rules based on service-deployment-mode for a given
    environment."""
    try:
        session = get_aws_session_or_abort()
        update_aws = UpdateALBRules(session)
        update_aws.update_alb_rules(environment=env)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))



from dataclasses import dataclass
from dataclasses import field
from typing import Literal
from typing import Optional
from typing import Union

import boto3

from dbt_platform_helper.platform_exception import PlatformException


@dataclass
class Parameter:

    name: str
    value: str
    arn: str = None
    data_type: Literal["text", "aws:ec2:image"] = "text"
    type: Literal["String", "StringList", "SecureString"] = (
        "SecureString"  # Returned as 'Type' from AWS
    )
    version: int = None
    tags: Optional[dict[str, str]] = field(default_factory=dict)

    def tags_to_list(self):
        return [{"Key": tag, "Value": value} for tag, value in self.tags.items()]

    def __str__(self):
        output = f"Application {self.name} with"

        return f"{output} no environments"

    def __eq__(self, other: "Parameter"):
        return str(self.name) == str(other.name)


class ParameterStore:
    def __init__(self, ssm_client: boto3.client, with_model: bool = False):
        self.ssm_client = ssm_client
        self.with_model = with_model

    def __fetch_tags(self, parameter: Parameter, normalise=True):
        response = self.ssm_client.list_tags_for_resource(
            ResourceType="Parameter", ResourceId=parameter.name
        )["TagList"]

        if normalise:
            return {tag["Key"]: tag["Value"] for tag in response}
        else:
            return response

    def get_ssm_parameter_by_name(
        self, parameter_name: str, add_tags: bool = False
    ) -> Union[dict, Parameter]:
        """
        Retrieves the latest version of a parameter from parameter store for a
        given name/arn.

        Args:
            parameter_name (str): The parameter name to retrieve the parameter value for.
            add_tags (bool): Whether to retrieve the tags for the SSM parameters requested
        Returns:
            dict: A dictionary representation of your ssm parameter
        """
        parameter = self.ssm_client.get_parameter(Name=parameter_name, WithDecryption=True)[
            "Parameter"
        ]

        if not self.with_model:
            return parameter

        model = Parameter(
            name=parameter["Name"],
            value=parameter["Value"],
            arn=parameter["ARN"],
            data_type=parameter["DataType"],
            type=parameter["Type"],
            version=parameter["Version"],
        )

        if add_tags:
            model.tags = self.__fetch_tags(model)

        return model

    def get_ssm_parameters_by_path(
        self, path: str, add_tags: bool = False
    ) -> Union[list[dict], list[Parameter]]:
        """
        Retrieves all SSM parameters for a given path from parameter store.

        Args:
            path (str): The parameter path to retrieve the parameters for. e.g. /copilot/applications/
            add_tags (bool): Whether to retrieve the tags for the SSM parameters requested
        Returns:
            list: A list of dictionaries containing all SSM parameters under the provided path.
        """

        parameters = []
        paginator = self.ssm_client.get_paginator("get_parameters_by_path")
        page_iterator = paginator.paginate(Path=path, Recursive=True, WithDecryption=True)

        for page in page_iterator:
            parameters.extend(page.get("Parameters", []))

        if not self.with_model:
            if parameters:
                return parameters
            else:
                raise ParameterNotFoundForPathException()

        to_model = lambda parameter: Parameter(
            name=parameter["Name"],
            value=parameter["Value"],
            arn=parameter["ARN"],
            data_type=parameter["DataType"],
            type=parameter["Type"],
            version=parameter["Version"],
        )

        models = [to_model(param) for param in parameters]

        if add_tags:
            for model in models:
                model.tags = self.__fetch_tags(model)

        return models

    def put_parameter(self, data_dict: dict) -> dict:
        return self.ssm_client.put_parameter(**data_dict)


class ParameterNotFoundForPathException(PlatformException):
    """Exception raised when no parameters are found for a given path."""



import random
import string
import subprocess
from typing import Any
from typing import Optional

from botocore.exceptions import ClientError

from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.platform_exception import ValidationException
from dbt_platform_helper.providers.vpc import Vpc
from dbt_platform_helper.utilities.decorators import retry
from dbt_platform_helper.utilities.decorators import wait_until


class ECSException(PlatformException):
    pass


class ECSAgentNotRunningException(ECSException):
    def __init__(self):
        super().__init__("""ECS exec agent never reached "RUNNING" status""")


class NoClusterException(ECSException):
    def __init__(self, application_name: str, environment: str):
        super().__init__(
            f"""No ECS cluster found for "{application_name}" in "{environment}" environment."""
        )


class ECS:
    def __init__(self, ecs_client, ssm_client, application_name: str, env: str):
        self.ecs_client = ecs_client
        self.ssm_client = ssm_client
        self.application_name = application_name
        self.env = env

    def start_ecs_task(
        self,
        cluster_name: str,
        container_name: str,
        task_def_arn: str,
        vpc_config: Vpc,
        env_vars: list[dict] = None,
    ):
        container_override = {"name": container_name}
        if env_vars:
            container_override["environment"] = env_vars

        response = self.ecs_client.run_task(
            taskDefinition=task_def_arn,
            cluster=cluster_name,
            capacityProviderStrategy=[
                {"capacityProvider": "FARGATE", "weight": 1, "base": 0},
            ],
            enableExecuteCommand=True,
            networkConfiguration={
                "awsvpcConfiguration": {
                    "subnets": vpc_config.public_subnets,
                    "securityGroups": vpc_config.security_groups,
                    "assignPublicIp": "ENABLED",
                }
            },
            overrides={"containerOverrides": [container_override]},
        )

        return response.get("tasks", [{}])[0].get("taskArn")

    def get_cluster_arn_by_name(self, cluster_name: str) -> str:
        clusters = self.ecs_client.describe_clusters(
            clusters=[
                cluster_name,
            ],
        )["clusters"]
        if len(clusters) == 1 and "clusterArn" in clusters[0]:
            return clusters[0]["clusterArn"]

        raise NoClusterException(self.application_name, self.env)

    def get_cluster_arn_by_copilot_tag(self) -> str:
        """Returns the ARN of the ECS cluster for the given application and
        environment."""
        for cluster_arn in self.ecs_client.list_clusters()["clusterArns"]:
            tags_response = self.ecs_client.list_tags_for_resource(resourceArn=cluster_arn)
            tags = tags_response["tags"]

            app_key_found = False
            env_key_found = False
            cluster_key_found = False

            for tag in tags:
                if tag["key"] == "copilot-application" and tag["value"] == self.application_name:
                    app_key_found = True
                if tag["key"] == "copilot-environment" and tag["value"] == self.env:
                    env_key_found = True
                if tag["key"] == "aws:cloudformation:logical-id" and tag["value"] == "Cluster":
                    cluster_key_found = True

            if app_key_found and env_key_found and cluster_key_found:
                return cluster_arn

        raise NoClusterException(self.application_name, self.env)

    def get_or_create_task_name(self, addon_name: str, parameter_name: str) -> str:
        """Fetches the task name from SSM or creates a new one if not found."""
        try:
            return self.ssm_client.get_parameter(Name=parameter_name)["Parameter"]["Value"]
        except self.ssm_client.exceptions.ParameterNotFound:
            random_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=12))
            return f"conduit-{self.application_name}-{self.env}-{addon_name}-{random_id}"

    def get_ecs_task_arns(
        self,
        cluster: str,
        max_results: int = 100,
        desired_status: str = "RUNNING",
        service_name: Optional[str] = None,
        started_by: Optional[str] = None,
        task_def_family: Optional[str] = None,
    ) -> list[str]:
        """Returns the ECS task ARNs based on the parameters provided."""

        params = {
            "cluster": cluster,
            "maxResults": max_results,
            "desiredStatus": desired_status,
        }

        if service_name:
            params["serviceName"] = service_name
        if started_by:
            params["startedBy"] = started_by
        if task_def_family:
            params["family"] = task_def_family

        tasks = self.ecs_client.list_tasks(**params)

        if not tasks["taskArns"]:
            return []

        return tasks["taskArns"]

    @retry()
    def exec_task(self, cluster_arn: str, task_arn: str, subprocess_call=subprocess.call):
        result = subprocess_call(
            f"aws ecs execute-command --cluster {cluster_arn} "
            f"--task {task_arn} "
            f"--interactive --command bash ",
            shell=True,
        )
        if result != 0:
            raise PlatformException("Failed to exec into ECS task.")
        return result

    @wait_until(
        max_attempts=25,
        exceptions_to_catch=(ECSException,),
        message_on_false="ECS Agent Not running",
    )
    def ecs_exec_is_available(self, cluster_arn: str, task_arns: list[str]) -> bool:
        """
        Checks if the ExecuteCommandAgent is running on the specified ECS task.

        Waits for up to 25 attempts, then raises ECSAgentNotRunning if still not
        running.
        """
        if not task_arns:
            raise ValidationException("No task ARNs provided")
        task_details = self.ecs_client.describe_tasks(cluster=cluster_arn, tasks=task_arns)

        if not task_details["tasks"]:
            raise ECSException("No ECS tasks returned.")
        container_details = task_details["tasks"][0]["containers"][0]
        if container_details.get("managedAgents", None):
            managed_agents = container_details["managedAgents"]
        else:
            raise ECSException("No managed agent on ecs task.")

        execute_command_agent = [
            agent for agent in managed_agents if agent["name"] == "ExecuteCommandAgent"
        ]
        if not execute_command_agent:
            raise ECSException("No ExecuteCommandAgent on ecs task.")
        return execute_command_agent[0]["lastStatus"] == "RUNNING"

    @wait_until(
        max_attempts=20,
        message_on_false="ECS task did not register in time",
    )
    def wait_for_task_to_register(self, cluster_arn: str, task_family: str) -> list[str]:
        task_arns = self.get_ecs_task_arns(cluster=cluster_arn, task_def_family=task_family)
        if task_arns:
            return task_arns
        return False

    def get_service_deployment_state(
        self, cluster_name: str, service_name: str, start_time: float
    ) -> tuple[Optional[str], Optional[str]]:
        """
        Returns status & statusReason for the deployment of an ECS service.

        statusReason can be:
            PENDING | SUCCESSFUL | STOPPED | STOP_REQUESTED |
            IN_PROGRESS | ROLLBACK_REQUESTED | ROLLBACK_IN_PROGRESS |
            ROLLBACK_SUCCESSFUL | ROLLBACK_FAILED
        """
        resp = self.ecs_client.list_service_deployments(
            cluster=cluster_name, service=service_name, createdAt={"after": start_time}
        )
        deployments = resp.get("serviceDeployments", [])

        if not deployments:
            return None, f"No deployments found for '{service_name}'"

        return deployments[0].get("status"), deployments[0].get("statusReason")

    def get_container_names_from_ecs_tasks(
        self, cluster_name: str, task_ids: list[str]
    ) -> list[str]:
        """Retrieve container names from each ECS task provided."""

        response = self.ecs_client.describe_tasks(cluster=cluster_name, tasks=task_ids)

        names = []
        for task in response.get("tasks", []):
            for container in task.get("containers", []):
                if container["name"] not in names:
                    names.append(container["name"])
        return names

    def register_task_definition(
        self,
        application: str,
        environment: str,
        service: str,
        task_definition: dict,
        image_tag: Optional[str] = None,
    ) -> str:
        """Register a new task definition revision using provided model and
        containerDefinitions."""

        for container in task_definition["containerDefinitions"]:
            if container["name"] == service:

                # Append tag to the image URI
                container["image"] = f"{container['image']}:{image_tag}"

                # Add DataDog Docker labels https://docs.datadoghq.com/getting_started/tagging/unified_service_tagging/?tab=ecs#partial-configuration
                container["dockerLabels"] = {
                    "com.datadoghq.tags.env": environment,
                    "com.datadoghq.tags.service": f"{application}-{service}",
                    "com.datadoghq.tags.version": image_tag,
                }
                break

        try:
            task_definition_response = self.ecs_client.register_task_definition(**task_definition)
            return task_definition_response["taskDefinition"]["taskDefinitionArn"]
        except ClientError as err:
            raise PlatformException(f"Error registering task definition: {err}")

    def update_service(
        self,
        service: str,
        task_def_arn: str,
        environment: str,
        application: str,
        desired_count: int,
    ) -> dict[str, Any]:
        """Update an ECS service and return the response."""

        try:
            service_response = self.ecs_client.update_service(
                cluster=f"{application}-{environment}-cluster",
                service=f"{application}-{environment}-{service}",
                taskDefinition=task_def_arn,
                desiredCount=desired_count,
            )
            return service_response["service"]
        except ClientError as err:
            raise PlatformException(f"Error updating ECS service: {err}")

    def describe_service(self, service: str, environment: str, application: str) -> dict[str, Any]:
        """Return information about an ECS service."""

        try:
            service_response = self.ecs_client.describe_services(
                cluster=f"{application}-{environment}-cluster",
                services=[
                    f"{application}-{environment}-{service}",
                ],
            )
            return service_response["services"][0]
        except ClientError as err:
            raise PlatformException(f"Error retrieving ECS service: {err}")

    def describe_tasks(self, cluster_name: str, task_ids: list[str]) -> list[dict[str, Any]]:
        """Return information about ECS tasks."""

        try:
            response = self.ecs_client.describe_tasks(cluster=cluster_name, tasks=task_ids)
            return response["tasks"]
        except ClientError as err:
            raise PlatformException(f"Error retrieving ECS tasks: {err}")



import json
from typing import Dict
from typing import List

from boto3 import Session

from dbt_platform_helper.constants import MANAGED_BY_PLATFORM_TERRAFORM
from dbt_platform_helper.constants import ROUTED_TO_PLATFORM_MODES
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.parameter_store import ParameterStore
from dbt_platform_helper.utils.aws import get_aws_session_or_abort


def normalise_to_cidr(ip: str):
    if "/" in ip:
        return ip
    SINGLE_IPV4_CIDR_PREFIX_LENGTH = "32"
    return f"{ip}/{SINGLE_IPV4_CIDR_PREFIX_LENGTH}"


class ALBDataNormaliser:

    @staticmethod
    def tags_to_dict(tags: List[Dict[str, str]]) -> Dict[str, str]:
        return {tag.get("Key", ""): tag.get("Value", "") for tag in tags}

    @staticmethod
    def conditions_to_dict(conditions: List[Dict[str, List[str]]]) -> Dict[str, List[str]]:
        return {condition.get("Field", ""): condition.get("Values", "") for condition in conditions}


class LoadBalancerProvider:

    def __init__(self, session: Session = None, io: ClickIOProvider = ClickIOProvider()):
        self.session = session
        self.evlb_client = self._get_client("elbv2")
        self.rg_tagging_client = self._get_client("resourcegroupstaggingapi")
        self.parameter_store_provider = ParameterStore(self._get_client("ssm"))
        self.io = io

    def _get_client(self, client: str):
        if not self.session:
            self.session = get_aws_session_or_abort()
        return self.session.client(client)

    def find_target_group(self, app: str, env: str, svc: str) -> str:

        # TODO once copilot is gone this is no longer needed
        try:
            result = self.parameter_store_provider.get_ssm_parameter_by_name(
                f"/platform/applications/{app}/environments/{env}"
            )["Value"]
            env_config = json.loads(result)
            service_deployment_mode = env_config["service_deployment_mode"]
        except Exception:
            service_deployment_mode = "copilot"

        if service_deployment_mode in ROUTED_TO_PLATFORM_MODES:
            application_key = "application"
            environment_key = "environment"
            service_key = "service"
        else:
            application_key = "copilot-application"
            environment_key = "copilot-environment"
            service_key = "copilot-service"
        target_group_arn = None

        paginator = self.rg_tagging_client.get_paginator("get_resources")
        page_iterator = paginator.paginate(
            TagFilters=[
                {
                    "Key": application_key,
                    "Values": [
                        app,
                    ],
                },
                {
                    "Key": environment_key,
                    "Values": [
                        env,
                    ],
                },
                {
                    "Key": service_key,
                    "Values": [
                        svc,
                    ],
                },
            ],
            ResourceTypeFilters=[
                "elasticloadbalancing:targetgroup",
            ],
        )

        for page in page_iterator:
            for resource in page["ResourceTagMappingList"]:
                tags = {tag["Key"]: tag["Value"] for tag in resource["Tags"]}

                if (
                    tags.get(service_key) == svc
                    and tags.get(environment_key) == env
                    and tags.get(application_key) == app
                ):
                    target_group_arn = resource["ResourceARN"]

        if not target_group_arn:
            self.io.error(
                f"No target group found for application: {app}, environment: {env}, service: {svc}",
            )

        return target_group_arn

    def get_target_groups(self, target_group_arns: List[str]) -> List[dict]:
        tgs = []
        paginator = self.evlb_client.get_paginator("describe_target_groups")
        page_iterator = paginator.paginate(TargetGroupArns=target_group_arns)
        for page in page_iterator:
            tgs.extend(page["TargetGroups"])

        return tgs

    def get_target_groups_with_tags(
        self, target_group_arns: List[str], normalise: bool = True
    ) -> List[dict]:
        target_groups = self.get_target_groups(target_group_arns)

        tags = self.get_resources_tag_descriptions(target_groups, "TargetGroupArn")

        tgs_with_tags = self.merge_in_tags_by_resource_arn(target_groups, tags, "TargetGroupArn")

        if normalise:
            for tg in tgs_with_tags:
                tg["Tags"] = ALBDataNormaliser.tags_to_dict(tg["Tags"])
        return tgs_with_tags

    def get_https_certificate_for_listener(self, listener_arn: str, env: str):
        certificates = []
        paginator = self.evlb_client.get_paginator("describe_listener_certificates")
        page_iterator = paginator.paginate(ListenerArn=listener_arn)
        for page in page_iterator:
            certificates.extend(page["Certificates"])

        try:
            certificate_arn = next(c["CertificateArn"] for c in certificates if c["IsDefault"])
        except StopIteration:
            raise CertificateNotFoundException(env)

        return certificate_arn

    def get_https_certificate_for_application(self, app: str, env: str) -> str:
        listener_arn = self.get_https_listener_for_application(app, env)
        return self.get_https_certificate_for_listener(listener_arn, env)

    def get_listeners_for_load_balancer(self, load_balancer_arn: str) -> List[dict]:
        listeners = []
        paginator = self.evlb_client.get_paginator("describe_listeners")
        page_iterator = paginator.paginate(LoadBalancerArn=load_balancer_arn)
        for page in page_iterator:
            listeners.extend(page["Listeners"])

        return listeners

    def get_https_listener_for_application(self, app: str, env: str) -> str:
        load_balancer_arn = self.get_load_balancer_for_application(app, env)
        self.io.debug(f"Load Balancer ARN: {load_balancer_arn}")
        listeners = self.get_listeners_for_load_balancer(load_balancer_arn)

        listener_arn = None

        try:
            listener_arn = next(l["ListenerArn"] for l in listeners if l["Protocol"] == "HTTPS")
        except StopIteration:
            pass

        if not listener_arn:
            raise ListenerNotFoundException(app, env)

        return listener_arn

    def get_load_balancers(self) -> List[dict]:
        load_balancers = []
        paginator = self.evlb_client.get_paginator("describe_load_balancers")
        page_iterator = paginator.paginate()
        for page in page_iterator:
            load_balancers.extend(lb["LoadBalancerArn"] for lb in page["LoadBalancers"])

        return load_balancers

    def get_load_balancer_for_application(self, app: str, env: str) -> str:
        load_balancers = self.get_load_balancers()
        tag_descriptions = []
        for i in range(0, len(load_balancers), 20):
            chunk = load_balancers[i : i + 20]
            tag_descriptions.extend(
                self.evlb_client.describe_tags(ResourceArns=chunk)["TagDescriptions"]
            )  # describe_tags cannot be paginated - 04/04/2025

        for lb in tag_descriptions:
            tags = {t["Key"]: t["Value"] for t in lb["Tags"]}
            # TODO: DBTP-1967: copilot hangover, creates coupling to specific tags could update to check application and environment
            if (
                tags.get("copilot-application") == app
                and tags.get("copilot-environment") == env
                and tags.get("managed-by", "") == MANAGED_BY_PLATFORM_TERRAFORM
            ):
                return lb["ResourceArn"]

        raise LoadBalancerNotFoundException(app, env)

    def get_host_header_conditions(self, listener_arn: str, target_group_arn: str) -> list:
        rules = []
        paginator = self.evlb_client.get_paginator("describe_rules")
        page_iterator = paginator.paginate(ListenerArn=listener_arn)
        for page in page_iterator:
            rules.extend(page["Rules"])

        conditions = []

        for rule in rules:
            for action in rule["Actions"]:
                if "TargetGroupArn" in action:
                    if action["Type"] == "forward" and action["TargetGroupArn"] == target_group_arn:
                        conditions = rule["Conditions"]

        if not conditions:
            raise ListenerRuleConditionsNotFoundException(listener_arn)

        # filter to host-header conditions
        conditions = [
            {i: condition[i] for i in condition if i != "Values"}
            for condition in conditions
            if condition["Field"] == "host-header"
        ]

        # remove internal hosts
        conditions[0]["HostHeaderConfig"]["Values"] = [
            v for v in conditions[0]["HostHeaderConfig"]["Values"]
        ]

        return conditions

    def get_rules_tag_descriptions_by_listener_arn(self, listener_arn: str) -> list:
        rules = self.get_listener_rules_by_listener_arn(listener_arn)

        return self.get_resources_tag_descriptions(rules)

    def merge_in_tags_by_resource_arn(
        self,
        resources: List[dict],
        tag_descriptions: List[dict],
        resources_identifier: str = "RuleArn",
    ):
        tags_by_resource_arn = {
            rule_tags.get("ResourceArn"): rule_tags for rule_tags in tag_descriptions if rule_tags
        }
        for resource in resources:
            tags = tags_by_resource_arn[resource[resources_identifier]]
            resource.update(tags)
        return resources

    def get_rules_with_tags_by_listener_arn(
        self, listener_arn: str, normalise: bool = True
    ) -> list:
        rules = self.get_listener_rules_by_listener_arn(listener_arn)

        tags = self.get_resources_tag_descriptions(rules)

        rules_with_tags = self.merge_in_tags_by_resource_arn(rules, tags)

        if normalise:
            for rule in rules_with_tags:
                rule["Conditions"] = ALBDataNormaliser.conditions_to_dict(rule["Conditions"])
                rule["Tags"] = ALBDataNormaliser.tags_to_dict(rule["Tags"])

        return rules_with_tags

    def get_listener_rules_by_listener_arn(self, listener_arn: str) -> list:
        rules = []
        paginator = self.evlb_client.get_paginator("describe_rules")
        page_iterator = paginator.paginate(ListenerArn=listener_arn)
        for page in page_iterator:
            rules.extend(page["Rules"])

        return rules

    def get_resources_tag_descriptions(
        self, resources: list, resource_identifier: str = "RuleArn"
    ) -> list:
        tag_descriptions = []
        chunk_size = 20

        for i in range(0, len(resources), chunk_size):
            chunk = resources[i : i + chunk_size]
            resource_arns = [r[resource_identifier] for r in chunk]
            response = self.evlb_client.describe_tags(
                ResourceArns=resource_arns
            )  # describe_tags cannot be paginated - 04/04/2025
            tag_descriptions.extend(response["TagDescriptions"])

        return tag_descriptions

    def create_rule(
        self,
        listener_arn: str,
        actions: list,
        conditions: list,
        priority: int,
        tags: list,
    ):
        return self.evlb_client.create_rule(
            ListenerArn=listener_arn,
            Priority=priority,
            Conditions=conditions,
            Actions=actions,
            Tags=tags,
        )

    def create_forward_rule(
        self,
        listener_arn: str,
        target_group_arn: str,
        rule_name: str,
        priority: int,
        conditions: list,
        additional_tags: list = [],
    ):
        return self.create_rule(
            listener_arn=listener_arn,
            priority=priority,
            conditions=conditions,
            actions=[{"Type": "forward", "TargetGroupArn": target_group_arn}],
            tags=[{"Key": "name", "Value": rule_name}, *additional_tags],
        )

    def create_header_rule(
        self,
        listener_arn: str,
        target_group_arn: str,
        header_name: str,
        values: list,
        rule_name: str,
        priority: int,
        conditions: list,
        additional_tags: list = [],
    ):

        combined_conditions = [
            {
                "Field": "http-header",
                "HttpHeaderConfig": {"HttpHeaderName": header_name, "Values": values},
            }
        ] + conditions

        self.create_forward_rule(
            listener_arn,
            target_group_arn,
            rule_name,
            priority,
            combined_conditions,
            additional_tags,
        )

        self.io.debug(
            f"Creating listener rule {rule_name} for HTTPS Listener with arn {listener_arn}.\nIf request header {header_name} contains one of the values {values}, the request will be forwarded to target group with arn {target_group_arn}.\n\n",
        )

    def create_source_ip_rule(
        self,
        listener_arn: str,
        target_group_arn: str,
        values: list,
        rule_name: str,
        priority: int,
        conditions: list,
        additional_tags: list = [],
    ):
        combined_conditions = [
            {
                "Field": "source-ip",
                "SourceIpConfig": {"Values": [normalise_to_cidr(value) for value in values]},
            }
        ] + conditions

        self.create_forward_rule(
            listener_arn,
            target_group_arn,
            rule_name,
            priority,
            combined_conditions,
            additional_tags,
        )

        self.io.debug(
            f"Creating listener rule {rule_name} for HTTPS Listener with arn {listener_arn}.\nIf request source ip matches one of the values {values}, the request will be forwarded to target group with arn {target_group_arn}.\n\n",
        )

    def delete_listener_rule_by_tags(self, tag_descriptions: list, tag_name: str) -> list:
        deleted_rules = []

        for description in tag_descriptions:
            tags = {t["Key"]: t["Value"] for t in description["Tags"]}
            if tags.get("name") == tag_name:
                if description["ResourceArn"]:
                    self.evlb_client.delete_rule(RuleArn=description["ResourceArn"])
                    deleted_rules.append(description)

        return deleted_rules

    def delete_listener_rule_by_resource_arn(self, resource_arn: str) -> list:
        return self.evlb_client.delete_rule(RuleArn=resource_arn)


class LoadBalancerException(PlatformException):
    pass


class LoadBalancerNotFoundException(LoadBalancerException):
    def __init__(self, application_name, env):
        super().__init__(
            f"No load balancer found for environment {env} in the application {application_name}."
        )


class ListenerNotFoundException(LoadBalancerException):
    def __init__(self, application_name, env):
        super().__init__(
            f"No HTTPS listener found for environment {env} in the application {application_name}."
        )


class ListenerRuleNotFoundException(LoadBalancerException):
    pass


class ListenerRuleConditionsNotFoundException(LoadBalancerException):
    def __init__(self, listener_arn):
        super().__init__(f"No listener rule conditions found for listener ARN: {listener_arn}")


class CertificateNotFoundException(PlatformException):
    def __init__(self, environment_name: str):
        super().__init__(
            f"""No certificate found with domain name matching environment {environment_name}."."""
        )



def deep_merge(base, override):
    result = base.copy()

    for k, v in override.items():
        if k in result and isinstance(result[k], dict) and isinstance(v, dict):
            result[k] = deep_merge(result[k], v)
        else:
            result[k] = v

    return result



import boto3
from botocore.exceptions import ClientError

from dbt_platform_helper.platform_exception import PlatformException


class S3Provider:

    def __init__(self, client: boto3.client):
        self.client = client

    def get_object(self, bucket_name: str, object_key: str) -> str:
        """Returns an object from an S3 bucket."""

        try:
            content = self.client.get_object(Bucket=bucket_name, Key=object_key)
            return content["Body"].read().decode("utf-8")
        except ClientError as e:
            raise PlatformException(
                f"Failed to get '{object_key}' from '{bucket_name}'. Error: {e}"
            )



from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.terraform_manifest import TerraformManifestProvider


class TerraformEnvironmentException(PlatformException):
    pass


class EnvironmentNotFoundException(TerraformEnvironmentException):
    pass


class TerraformEnvironment:
    def __init__(
        self,
        config_provider,
        manifest_provider: TerraformManifestProvider = None,
        io: ClickIOProvider = ClickIOProvider(),
        platform_helper_versioning: PlatformHelperVersioning = None,
    ):
        self.io = io
        self.config_provider = config_provider
        self.manifest_provider = manifest_provider or TerraformManifestProvider()
        self.platform_helper_versioning = platform_helper_versioning

    def generate(
        self,
        environment_name: str,
    ):

        self.platform_helper_versioning.check_platform_helper_version_mismatch()

        config = self.config_provider.get_enriched_config()

        if environment_name not in config.get("environments").keys():
            raise EnvironmentNotFoundException(
                f"cannot generate terraform for environment {environment_name}.  It does not exist in your configuration"
            )

        self.manifest_provider.generate_environment_config(
            config,
            environment_name,
            self.platform_helper_versioning.get_template_version(),
            self.platform_helper_versioning.get_extensions_module_source(),
        )



#!/usr/bin/env python

# application commands are deprecated, do not spend time refactoring them
# Service teams are trained to use them as a replacement for cf app(s)

import time
from datetime import datetime
from datetime import timedelta

import click
from prettytable import PrettyTable

from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.click import ClickDocOptGroup

YELLOW = "\033[93m"
CYAN = "\033[96m"


def get_query_results(env, app, query_string, timeout):
    label_text = click.style("Waiting for results:", fg="yellow")
    fill_char = click.style("#", fg="yellow")
    empty_char = click.style("-", fg="yellow", dim=True)

    application = load_application(app)
    project_session = application.environments[env].session

    click.secho(
        f"Showing status for app {app}",
        fg="green",
    )
    logs_client = project_session.client("logs")
    ecs_client = project_session.client("ecs")

    response = ecs_client.list_clusters()

    # Find clusters log-group
    for cluster in response["clusterArns"]:
        cluster_name = cluster.split("/")[-1]
        APP = cluster_name.split("-")[0]
        ENV = cluster_name.split("-")[1]
        if app == APP and env == ENV:
            log_group_name = f"/aws/ecs/containerinsights/{cluster_name}/performance"
            break

    # Container stats are 5 mins behind realtime.
    date_time = datetime.now()
    start_time = int((date_time - timedelta(minutes=5)).timestamp())
    end_time = int((date_time - timedelta(minutes=4)).timestamp())

    click.echo(
        click.style("Date & Time:  ", fg="cyan")
        + click.style(f"{datetime.utcfromtimestamp(start_time)}", fg="cyan", bold=True)
    )

    # First you need to create a log query, then retrieve the log query.
    cpu_response_id = logs_client.start_query(
        logGroupName=log_group_name,
        startTime=start_time,
        endTime=end_time,
        queryString=query_string,
    )

    elapsed = 0
    # Need to wait a few seconds for the query to be available, up to timeout value.
    with click.progressbar(
        label=label_text,
        length=timeout,
        show_eta=False,
        show_percent=False,
        fill_char=fill_char,
        empty_char=empty_char,
    ) as progress_bar:
        while (
            len(logs_client.get_query_results(queryId=cpu_response_id["queryId"])["results"]) == 0
            and elapsed < timeout
        ):
            time.sleep(1)
            progress_bar.update(1)
            elapsed = elapsed + 1

    if elapsed == timeout:
        click.secho(
            f"Timeout threshold breached, no results returned in {timeout} seconds.\nExiting...",
            fg="red",
        )
        exit()

    cpu_response = logs_client.get_query_results(queryId=cpu_response_id["queryId"])

    return cpu_response


@click.group(chain=True, cls=ClickDocOptGroup, deprecated=True)
def application():
    """[DEPRECATED] Application metrics."""
    PlatformHelperVersioning().check_if_needs_update()


@application.command(deprecated=True)
@click.option("--env", type=str, required=True)
@click.option("--app", type=str, required=True)
@click.option("--storage", is_flag=True)
@click.option("--network", is_flag=True)
def container_stats(env, app, storage, network):
    """[DEPRECATED] Command to get application container level metrics."""

    # Query string to get the required container stats
    query_string = "stats max(CpuUtilized), max(MemoryUtilized) by TaskId, ContainerName, TaskDefinitionFamily, TaskDefinitionRevision, Image, StorageReadBytes, StorageWriteBytes, NetworkRxPackets, NetworkTxBytes | filter Type='Container' | sort TaskId, ContainerName desc"
    cpu_response = get_query_results(env, app, query_string, 15)

    click.echo(
        click.style(f"\n{'Name:':<20}", fg="green") + click.style(f"{app}", fg="green", bold=True)
    )

    click.echo(
        click.style(f"{'Environment:':<20}", fg="green")
        + click.style(f"{env}", fg="green", bold=True)
    )

    click.echo(
        click.style(f"{'No of instances:':<20}", fg="green")
        + click.style(len(cpu_response["results"]), fg="green", bold=True)
    )

    index = 0
    for field in cpu_response["results"]:
        # Not all containers have an image not present, so need to pad.
        if len(field) == 10:
            field.insert(4, {"field": "Image", "value": ""})

        task = field[0]["value"]
        cont_name = field[1]["value"]
        cont_name_short = field[2]["value"].split("-")[-1]
        task_def_revision = field[3]["value"]
        image = field[4]["value"]
        storage_read = field[5]["value"]
        storage_write = field[6]["value"]
        network_read = field[7]["value"]
        network_write = field[8]["value"]
        cpu = f"{float(field[9]['value']):.1f}%"
        memory = f"{field[10]['value']}M"

        # Nothing to compare to at start.
        if index > 0:
            end_index = index - 1
        else:
            end_index = 0

        # If its a new task then create new table and display the headings.
        if (
            index == 0
            or cpu_response["results"][index][0]["value"]
            != cpu_response["results"][end_index][0]["value"]
        ):
            result_table = PrettyTable()
            click.echo(
                click.style(f"\n{'Type:':<10}", fg="green")
                + click.style(f"{cont_name_short}", fg="green", bold=True)
            )

            click.echo(
                click.style(f"{'Task ID:':<10}", fg="green")
                + click.style(f"{task}", fg="green", bold=True)
                + "\n"
            )

            heading_fields = [
                f"{CYAN}Container Name",
                f"{CYAN}CPU",
                f"{CYAN}Memory",
                f"{CYAN}Revision",
            ]

            # Optional parameters.
            if storage:
                heading_fields.append(f"{CYAN}Disk Read")
                heading_fields.append(f"{CYAN}Disk Write")
            if network:
                heading_fields.append(f"{CYAN}Net Read")
                heading_fields.append(f"{CYAN}Net Write")

            heading_fields.append(f"{CYAN}Image")

            result_table.field_names = heading_fields
            for item in heading_fields:
                result_table.align[f"{item}"] = "l"
            result_table.border = False

        # Print container stats
        values = [
            f"{YELLOW}{cont_name}",
            f"{YELLOW}{cpu}",
            f"{YELLOW}{memory}",
            f"{YELLOW}{task_def_revision}",
        ]

        # Optional stats.
        if storage:
            values.append(f"{storage_read}")
            values.append(f"{storage_write}")
        if network:
            values.append(f"{network_read}")
            values.append(f"{network_write}")
        values.append(f"{image}")

        result_table.add_row(values)

        # Print table when new task is next or at end of loop.
        if (
            index == len(cpu_response["results"]) - 1
            or cpu_response["results"][index][0]["value"]
            != cpu_response["results"][index + 1][0]["value"]
        ):
            click.secho(result_table)

        index = index + 1


@application.command(deprecated=True)
@click.option("--env", type=str, required=True)
@click.option("--app", type=str, required=True)
@click.option("--disk", is_flag=True)
@click.option("--storage", is_flag=True)
@click.option("--network", is_flag=True)
def task_stats(env, app, disk, storage, network):
    """[DEPRECATED] Command to get application task level metrics."""

    # Query string to get the required container stats
    query_string = "stats max(CpuUtilized), max(MemoryUtilized), max(EphemeralStorageUtilized) by TaskId, TaskDefinitionFamily, TaskDefinitionRevision, StorageReadBytes, StorageWriteBytes, NetworkRxPackets, NetworkTxBytes, KnownStatus | filter Type='Task' | sort TaskId desc"
    cpu_response = get_query_results(env, app, query_string, 15)

    click.echo(
        click.style(f"\n{'Name:':<20}", fg="green") + click.style(f"{app}", fg="green", bold=True)
    )

    click.echo(
        click.style(f"{'No of instances:':<20}", fg="green")
        + click.style(len(cpu_response["results"]), fg="green", bold=True)
        + "\n"
    )

    result_table = PrettyTable()
    heading_fields = [
        f"{CYAN}Type",
        f"{CYAN}TaskID",
        f"{CYAN}Revision",
        f"{CYAN}Status",
        f"{CYAN}CPU",
        f"{CYAN}Memory",
    ]

    # Add to heading if additional parameters are used.
    if disk:
        heading_fields.append(f"{CYAN}Disk")
    if storage:
        heading_fields.append(f"{CYAN}Disk Read")
        heading_fields.append(f"{CYAN}Disk Write")
    if network:
        heading_fields.append(f"{CYAN}Net Read")
        heading_fields.append(f"{CYAN}Net Write")

    result_table.field_names = heading_fields

    for item in heading_fields:
        result_table.align[f"{item}"] = "l"

    result_table.border = False

    for (
        task,
        taskdef,
        task_def_revision,
        storage_read,
        storage_write,
        network_read,
        network_write,
        status,
        cpu,
        memory,
        dsk,
    ) in cpu_response["results"]:
        cpu_value = f"{float(cpu['value']):.1f}" + "%"
        values = [
            f"{YELLOW}{taskdef['value'].split('-')[-1]}",
            f"{YELLOW}{task['value']}",
            f"{YELLOW}{task_def_revision['value']}",
            f"{YELLOW}{status['value']}",
            f"{YELLOW}{cpu_value}",
            f"{YELLOW}{memory['value'] + 'M'}",
        ]

        # Optional stats.
        if disk:
            values.append(f"{dsk['value'] + 'G'}")
        if storage:
            values.append(f"{storage_read['value']}")
            values.append(f"{storage_write['value']}")
        if network:
            values.append(f"{network_read['value']}")
            values.append(f"{network_write['value']}")

        result_table.add_row(values)

    click.secho(result_table)



import os

from dbt_platform_helper.constants import CODEBASE_PIPELINE_MODULE_PATH
from dbt_platform_helper.constants import ENVIRONMENT_PIPELINE_MODULE_PATH
from dbt_platform_helper.constants import PLATFORM_HELPER_VERSION_OVERRIDE_KEY
from dbt_platform_helper.constants import (
    TERRAFORM_CODEBASE_PIPELINES_MODULE_SOURCE_OVERRIDE_ENV_VAR,
)
from dbt_platform_helper.constants import (
    TERRAFORM_ENVIRONMENT_PIPELINES_MODULE_SOURCE_OVERRIDE_ENV_VAR,
)
from dbt_platform_helper.constants import (
    TERRAFORM_EXTENSIONS_MODULE_SOURCE_OVERRIDE_ENV_VAR,
)
from dbt_platform_helper.entities.semantic_version import (
    IncompatibleMajorVersionException,
)
from dbt_platform_helper.entities.semantic_version import (
    IncompatibleMinorVersionException,
)
from dbt_platform_helper.entities.semantic_version import SemanticVersion
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.environment_variable import (
    EnvironmentVariableProvider,
)
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.version import AWSCLIInstalledVersionProvider
from dbt_platform_helper.providers.version import CopilotInstalledVersionProvider
from dbt_platform_helper.providers.version import GithubLatestVersionProvider
from dbt_platform_helper.providers.version import InstalledVersionProvider
from dbt_platform_helper.providers.version import PyPiLatestVersionProvider
from dbt_platform_helper.providers.version import VersionProvider
from dbt_platform_helper.providers.version_status import VersionStatus


def running_as_installed_package():
    return "site-packages" in __file__


def allow_override_of_versioning_checks_fn():
    return not running_as_installed_package() or "PLATFORM_TOOLS_SKIP_VERSION_CHECK" in os.environ


class PlatformHelperVersionNotFoundException(PlatformException):
    def __init__(self, message=None):
        super().__init__(message or "Platform helper version could not be resolved.")


class PlatformHelperVersioning:
    def __init__(
        self,
        io: ClickIOProvider = ClickIOProvider(),
        config_provider: ConfigProvider = ConfigProvider(),
        environment_variable_provider: EnvironmentVariableProvider = EnvironmentVariableProvider(),
        latest_version_provider: VersionProvider = PyPiLatestVersionProvider,
        installed_version_provider: InstalledVersionProvider = InstalledVersionProvider(),
        allow_override_of_versioning_checks: bool = None,
        platform_helper_version_override: str = None,
    ):
        self.io = io
        self.config_provider = config_provider
        self.latest_version_provider = latest_version_provider
        self.installed_version_provider = installed_version_provider
        self.allow_override_of_versioning_checks = (
            allow_override_of_versioning_checks
            if allow_override_of_versioning_checks is not None
            else allow_override_of_versioning_checks_fn()
        )
        self.environment_variable_provider = environment_variable_provider
        self.platform_helper_version_override = platform_helper_version_override

    def is_auto(self):
        platform_config = self.config_provider.load_unvalidated_config_file()
        default_version = platform_config.get("default_versions", {}).get("platform-helper")
        return default_version == "auto"

    def get_required_version(self) -> str:
        if self.is_auto():
            return str(self.get_version_status().latest)
        else:
            return self.get_default_version()

    def _check_environment_is_configured_for_auto_versioning_within_a_pipeline(self):
        platform_helper_version_is_set_in_environment = self.environment_variable_provider.get(
            PLATFORM_HELPER_VERSION_OVERRIDE_KEY
        ) or self.environment_variable_provider.get("PLATFORM_HELPER_VERSION")
        modules_override_is_set_in_environment = (
            self.environment_variable_provider.get(
                TERRAFORM_EXTENSIONS_MODULE_SOURCE_OVERRIDE_ENV_VAR
            )
            or self.environment_variable_provider.get(
                TERRAFORM_CODEBASE_PIPELINES_MODULE_SOURCE_OVERRIDE_ENV_VAR
            )
            or self.environment_variable_provider.get(
                TERRAFORM_ENVIRONMENT_PIPELINES_MODULE_SOURCE_OVERRIDE_ENV_VAR
            )
        )
        if platform_helper_version_is_set_in_environment and modules_override_is_set_in_environment:
            return
        else:
            message = "You are on managed upgrades. Generate commands should only be running inside a pipeline environment."
            if self.allow_override_of_versioning_checks:
                self.io.warn(message)
                self.io.info("Bypassing versioning enforcement")
            else:
                self.io.abort_with_error(message)

    def check_platform_helper_version_mismatch(self):
        if self.is_auto():
            self._check_environment_is_configured_for_auto_versioning_within_a_pipeline()

        version_status = self.get_version_status()
        required_version = self.get_required_version()

        if SemanticVersion.is_semantic_version(required_version):
            required_version_semver = SemanticVersion.from_string(required_version)

            if not version_status.installed == required_version_semver:
                message = (
                    f"WARNING: You are running platform-helper v{version_status.installed} against "
                    f"v{required_version_semver} required by the project. Running anything besides the version required by the project may result in unpredictable and destructive changes."
                )
                self.io.warn(message)

    def check_if_needs_update(self):
        if self.allow_override_of_versioning_checks:
            return

        version_status = self.get_version_status()

        message = (
            f"You are running platform-helper v{version_status.installed}, upgrade to "
            f"v{version_status.latest} by running run `pip install "
            "--upgrade dbt-platform-helper`."
        )

        try:
            version_status.installed.validate_compatibility_with(version_status.latest)
        except IncompatibleMajorVersionException:
            self.io.error(message)
        except IncompatibleMinorVersionException:
            self.io.warn(message)

    def get_version_status(self) -> VersionStatus:
        locally_installed_version = self.installed_version_provider.get_semantic_version(
            "dbt-platform-helper"
        )

        latest_release = self.latest_version_provider.get_semantic_version("dbt-platform-helper")

        return VersionStatus(installed=locally_installed_version, latest=latest_release)

    def get_default_version(self):
        return (
            self.config_provider.load_unvalidated_config_file()
            .get("default_versions", {})
            .get("platform-helper")
        )

    def get_template_version(self):
        if self.is_auto():
            return self.environment_variable_provider.get(PLATFORM_HELPER_VERSION_OVERRIDE_KEY)
        if self.platform_helper_version_override:
            return self.platform_helper_version_override
        platform_helper_env_override = self.environment_variable_provider.get(
            PLATFORM_HELPER_VERSION_OVERRIDE_KEY
        )
        if platform_helper_env_override:
            return platform_helper_env_override

        return self.get_default_version()

    def get_pinned_version(self):
        if self.is_auto():
            return self.environment_variable_provider.get(PLATFORM_HELPER_VERSION_OVERRIDE_KEY)

        return None

    def _get_pipeline_modules_source(self, pipeline_module_path: str, override_env_var_key: str):
        pipeline_module_override = self.environment_variable_provider.get(override_env_var_key)

        if pipeline_module_override:
            return pipeline_module_override

        if self.platform_helper_version_override:
            return f"{pipeline_module_path}{self.platform_helper_version_override}"

        platform_helper_env_override = self.environment_variable_provider.get(
            PLATFORM_HELPER_VERSION_OVERRIDE_KEY
        )

        if platform_helper_env_override:
            return f"{pipeline_module_path}{platform_helper_env_override}"

        return f"{pipeline_module_path}{self.get_required_version()}"

    def get_environment_pipeline_modules_source(self):
        return self._get_pipeline_modules_source(
            ENVIRONMENT_PIPELINE_MODULE_PATH,
            TERRAFORM_ENVIRONMENT_PIPELINES_MODULE_SOURCE_OVERRIDE_ENV_VAR,
        )

    def get_codebase_pipeline_modules_source(self):
        return self._get_pipeline_modules_source(
            CODEBASE_PIPELINE_MODULE_PATH,
            TERRAFORM_CODEBASE_PIPELINES_MODULE_SOURCE_OVERRIDE_ENV_VAR,
        )

    def get_extensions_module_source(self):
        return self.environment_variable_provider.get(
            TERRAFORM_EXTENSIONS_MODULE_SOURCE_OVERRIDE_ENV_VAR
        )


class AWSVersioning:
    def __init__(
        self,
        latest_version_provider: VersionProvider = None,
        installed_version_provider: VersionProvider = None,
    ):
        self.latest_version_provider = latest_version_provider or GithubLatestVersionProvider
        self.installed_version_provider = (
            installed_version_provider or AWSCLIInstalledVersionProvider
        )

    def get_version_status(self) -> VersionStatus:
        return VersionStatus(
            self.installed_version_provider.get_semantic_version(),
            self.latest_version_provider.get_semantic_version("aws/aws-cli", True),
        )


class CopilotVersioning:
    def __init__(
        self,
        latest_version_provider: VersionProvider = None,
        installed_version_provider: VersionProvider = None,
    ):
        self.latest_version_provider = latest_version_provider or GithubLatestVersionProvider
        self.installed_version_provider = (
            installed_version_provider or CopilotInstalledVersionProvider
        )

    def get_version_status(self) -> VersionStatus:
        return VersionStatus(
            self.installed_version_provider.get_semantic_version(),
            self.latest_version_provider.get_semantic_version("aws/copilot-cli"),
        )



from pathlib import Path


class FileProvider:

    def load(path: str) -> str:
        pass

    @staticmethod
    def mkfile(base_path: str, file_name: str, contents: str, overwrite=False) -> str:
        file_path = Path(base_path).joinpath(file_name)
        file_exists = file_path.exists()
        if file_exists and not overwrite:
            return f"File {file_path} exists; doing nothing"

        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.write_text(contents)

        action = "overwritten" if file_exists and overwrite else "created"
        return f"File {file_path} {action}"

    @staticmethod
    def delete_file(base_path: str, file_name: str):
        file_path = Path(base_path) / file_name
        if file_path.exists():
            file_path.unlink()
            return f"{str(file_path)} has been deleted"



import click

from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.utils.click import ClickDocOptGroup


@click.group(chain=True, cls=ClickDocOptGroup)
def version():
    """Contains subcommands for getting version information about the current
    project."""


@version.command(help="Print the version of platform-tools required by the current project")
@click.option(
    "--pipeline",
    required=False,
    type=str,
    help="Take into account platform-tools version overrides in the specified pipeline",
)
def get_platform_helper_for_project(pipeline):
    """
    Deprecated: Required in pipelines for projects running platform-helper < 14.0.0.

    Version preferences removed, requires specifying platform-helper version in platform-config.yml
    """
    io = ClickIOProvider()
    try:
        io.info(PlatformHelperVersioning().get_required_version())
    except PlatformException as err:
        io.abort_with_error(str(err))



import functools
import time
from typing import Callable
from typing import Optional

from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider

SECONDS_BEFORE_RETRY = 3
RETRY_MAX_ATTEMPTS = 3


class RetryException(PlatformException):

    def __init__(
        self, function_name: str, max_attempts: int, original_exception: Optional[Exception] = None
    ):
        message = f"Function: {function_name} failed after {max_attempts} attempts"
        self.original_exception = original_exception
        if original_exception:
            message += f": \n{str(original_exception)}"
        super().__init__(message)


def retry(
    exceptions_to_catch: tuple = (Exception,),
    max_attempts: int = RETRY_MAX_ATTEMPTS,
    delay: float = SECONDS_BEFORE_RETRY,
    raise_custom_exception: bool = True,
    custom_exception: type = RetryException,
    io: ClickIOProvider = ClickIOProvider(),
):
    def decorator(func):
        func.__wrapped_by__ = "retry"

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            last_exception = None
            for attempt in range(max_attempts):
                try:
                    return func(*args, **kwargs)
                except exceptions_to_catch as e:
                    last_exception = e
                    io.debug(
                        f"Attempt {attempt+1}/{max_attempts} for {func.__name__} failed with exception {str(last_exception)}"
                    )
                    if attempt < max_attempts - 1:
                        time.sleep(delay)
            if raise_custom_exception:
                raise custom_exception(func.__name__, max_attempts, last_exception)
            raise last_exception

        return wrapper

    return decorator


def wait_until(
    exceptions_to_catch: tuple = (PlatformException,),
    max_attempts: int = RETRY_MAX_ATTEMPTS,
    delay: float = SECONDS_BEFORE_RETRY,
    raise_custom_exception: bool = True,
    custom_exception=RetryException,
    message_on_false="Condition not met",
    io: ClickIOProvider = ClickIOProvider(),
):
    """Wrap a function which returns a boolean."""

    def decorator(func: Callable[..., bool]):
        func.__wrapped_by__ = "wait_until"

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            last_exception = None
            for attempt in range(max_attempts):
                try:
                    result = func(*args, **kwargs)
                    if result:
                        return result
                    io.debug(
                        f"Attempt {attempt+1}/{max_attempts} for {func.__name__} returned falsy"
                    )
                except exceptions_to_catch as e:
                    last_exception = e
                    io.debug(
                        f"Attempt {attempt+1}/{max_attempts} for {func.__name__} failed with exception {str(last_exception)}"
                    )

                if attempt < max_attempts - 1:
                    time.sleep(delay)

            if not last_exception:  # If func returns false set last_exception
                last_exception = PlatformException(message_on_false)
            if (
                not raise_custom_exception
            ):  # Raise last_exception when you don't want custom exception
                raise last_exception
            else:
                raise custom_exception(func.__name__, max_attempts, last_exception)

        return wrapper

    return decorator



from collections import defaultdict
from pathlib import Path

from boto3 import Session

from dbt_platform_helper.domain.terraform_environment import (
    EnvironmentNotFoundException,
)
from dbt_platform_helper.providers.cloudformation import CloudFormation
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.files import FileProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.load_balancers import LoadBalancerProvider
from dbt_platform_helper.providers.vpc import Vpc
from dbt_platform_helper.providers.vpc import VpcNotFoundForNameException
from dbt_platform_helper.providers.vpc import VpcProvider
from dbt_platform_helper.utils.template import S3_CROSS_ACCOUNT_POLICY
from dbt_platform_helper.utils.template import camel_case
from dbt_platform_helper.utils.template import setup_templates


class CopilotEnvironment:
    def __init__(
        self,
        config_provider: ConfigProvider,
        vpc_provider: VpcProvider = None,
        cloudformation_provider: CloudFormation = None,
        session: Session = None,  # TODO: DBTP-1954: - this is a temporary fix, will fall away once _get_environment_vpc is updated.
        copilot_templating=None,
        io: ClickIOProvider = ClickIOProvider(),
        load_balancer_provider: LoadBalancerProvider = LoadBalancerProvider,
    ):
        self.config_provider = config_provider
        self.vpc_provider = vpc_provider
        self.copilot_templating = copilot_templating or CopilotTemplating(
            file_provider=FileProvider(),
        )
        self.io = io
        self.session = session
        self.load_balancer = load_balancer_provider(session)
        self.cloudformation_provider = cloudformation_provider

    def generate(self, environment_name: str) -> None:

        platform_config = self.config_provider.get_enriched_config()

        if environment_name not in platform_config.get("environments").keys():
            raise EnvironmentNotFoundException(
                f"Error: cannot generate copilot manifests for environment {environment_name}.  It does not exist in your configuration"
            )

        env_config = platform_config["environments"][environment_name]
        profile_for_environment = env_config.get("accounts", {}).get("deploy", {}).get("name")

        self.io.info(f"Using {profile_for_environment} for this AWS session")

        app_name = platform_config["application"]

        certificate_arn = self.load_balancer.get_https_certificate_for_application(
            app_name, environment_name
        )

        vpc = self._get_environment_vpc(
            self.session, app_name, environment_name, env_config.get("vpc", None)
        )

        copilot_environment_manifest = self.copilot_templating.generate_copilot_environment_manifest(
            environment_name=environment_name,
            # We need to correct the subnet id order before adding it to the template. See pydoc on below method for details.
            vpc=self._match_subnet_id_order_to_cloudformation_exports(environment_name, vpc),
            cert_arn=certificate_arn,
        )

        self.io.info(
            self.copilot_templating.write_environment_manifest(
                environment_name, copilot_environment_manifest
            )
        )

    # TODO: DBTP-1954: There should always be a vpc_name as defaults have been applied to the config.  This function can
    # probably fall away. We shouldn't need to check 3 different names (vpc_name, session.profile_name, {session.profile_name}-{env_name})
    # To be checked.
    def _get_environment_vpc(self, session: Session, app_name, env_name: str, vpc_name: str) -> Vpc:

        if not vpc_name:
            vpc_name = f"{session.profile_name}-{env_name}"

        try:
            vpc = self.vpc_provider.get_vpc(app_name, env_name, vpc_name)
        except VpcNotFoundForNameException:
            vpc = self.vpc_provider.get_vpc(app_name, env_name, session.profile_name)

        if not vpc:
            raise VpcNotFoundForNameException

        return vpc

    def _match_subnet_id_order_to_cloudformation_exports(
        self, environment_name: str, vpc: Vpc
    ) -> Vpc:
        """
        Addresses an issue identified in DBTP-1524 'If the order of the subnets
        in the environment manifest has changed, copilot env deploy tries to do
        destructive changes.'.

        Takes a Vpc object which has a private and public subnets attribute and
        sorts them to match the order within cfn exports.
        """

        exports = self.cloudformation_provider.get_cloudformation_exports_for_environment(
            environment_name
        )

        public_subnet_exports = []
        private_subnet_exports = []

        for export in exports:
            if export["Name"].endswith("-PublicSubnets"):
                public_subnet_exports = export["Value"].split(",")
            elif export["Name"].endswith("-PrivateSubnets"):
                private_subnet_exports = export["Value"].split(",")

        # If the elements match, regardless of order, use the list from the CloudFormation exports
        if set(vpc.public_subnets) == set(public_subnet_exports):
            vpc.public_subnets = public_subnet_exports
        if set(vpc.private_subnets) == set(private_subnet_exports):
            vpc.private_subnets = private_subnet_exports

        return vpc


class CopilotTemplating:
    def __init__(
        self,
        file_provider: FileProvider = FileProvider(),
        io: ClickIOProvider = ClickIOProvider(),
        # TODO: DBTP-1958: file_provider can be moved up a layer.  File writing can be the responsibility of CopilotEnvironment generate
        # Or we align with PlatformTerraformManifestGenerator and rename from Templating to reflect the file writing responsibility
    ):
        self.file_provider = file_provider
        self.templates = setup_templates()
        self.io = io

    def generate_copilot_environment_manifest(
        self, environment_name: str, vpc: Vpc, cert_arn: str
    ) -> str:
        env_template = self.templates.get_template("env/manifest.yml")

        return env_template.render(
            {
                "name": environment_name,
                "vpc_id": vpc.id,
                "pub_subnet_ids": vpc.public_subnets,
                "priv_subnet_ids": vpc.private_subnets,
                "certificate_arn": cert_arn,
            }
        )

    def write_environment_manifest(self, environment_name: str, manifest_contents: str) -> str:

        return self.file_provider.mkfile(
            ".",
            f"copilot/environments/{environment_name}/manifest.yml",
            manifest_contents,
            overwrite=True,
        )

    def generate_cross_account_s3_policies(self, environments: dict, extensions):
        resource_blocks = defaultdict(list)

        for ext_name, ext_data in extensions.items():
            for env_name, env_data in ext_data.get("environments", {}).items():
                if "cross_environment_service_access" in env_data:
                    bucket = env_data.get("bucket_name")
                    x_env_data = env_data["cross_environment_service_access"]
                    for access_name, access_data in x_env_data.items():
                        service = access_data.get("service")
                        read = access_data.get("read", False)
                        write = access_data.get("write", False)
                        if read or write:
                            resource_blocks[service].append(
                                {
                                    "bucket_name": bucket,
                                    "app_prefix": camel_case(f"{service}-{bucket}-{access_name}"),
                                    "bucket_env": env_name,
                                    "access_env": access_data.get("environment"),
                                    "bucket_account": environments.get(env_name, {})
                                    .get("accounts", {})
                                    .get("deploy", {})
                                    .get("id"),
                                    "read": read,
                                    "write": write,
                                }
                            )

        if not resource_blocks:
            self.io.info("\n>>> No cross-environment S3 policies to create.\n")
            return

        for service in sorted(resource_blocks.keys()):
            resources = resource_blocks[service]
            self.io.info(f"\n>>> Creating S3 cross account policies for {service}.\n")
            template = self.templates.get_template(S3_CROSS_ACCOUNT_POLICY)
            file_content = template.render({"resources": resources})
            output_dir = Path(".").absolute()
            file_path = f"copilot/{service}/addons/s3-cross-account-policy.yml"

            self.file_provider.mkfile(output_dir, file_path, file_content, True)
            self.io.info(f"File {file_path} created")
